-------------------------------------------------------------------------------
-------------------------------------------------------------------------------

-- this code takes care of storing and retrieving the data as packed eu-integers
-- a eu-byte = 0 to 214 and four of these can be packed into a 31-bit integer
-- which works out to 99.5% efficiency. Alrighty!

-------------------------------------------------------------------------------
-------------------------------------------------------------------------------

constant eu_char = 503,
        eu_char_high_boundary = floor(eu_char/2),
        eu_char_double = eu_char * eu_char,
        eu_char_triple = eu_char * eu_char * eu_char
-------------------------------------------------------------------------------
--DEV singlecolourbraces
-- option held in editor.ini...
integer singlecolourbraces
singlecolourbraces=1

function packChars(sequence four_eu_chars) -- pack 4 eu_chars into 31-bit integer
    atom a,b,c,d, n

    a=four_eu_chars[1]
    b=four_eu_chars[2]
    c=four_eu_chars[3]
    d=four_eu_chars[4]
    
    n = a + b*eu_char + c*eu_char_double
    if d > eu_char_high_boundary then
        d -= eu_char_high_boundary
        n += d*eu_char_triple
        n = 0-n
    else
        n += d*eu_char_triple
    end if
    return n
end function
-------------------------------------------------------------------------------
global function unpackChars(atom n) -- unpack eu-integer into 4 eu chars
    atom a, b, c, d
    if n < 0 then
        n = 0 - n -- change to positive
        d = eu_char_high_boundary
        d += floor(n/eu_char_triple)
        n = remainder(n, eu_char_triple)
    else
        d = floor(n/eu_char_triple)
        n = remainder(n, eu_char_triple)
    end if
    c = floor(n/eu_char_double)
    n = remainder(n, eu_char_double)
    b = floor(n/eu_char)
    a = remainder(n, eu_char)

    return {a,b,c,d}
end function
-------------------------------------------------------------------------------
global function AnalysedToPackedLine(sequence line)
-- takes a line of eu_chars and packs them into eu_integers with line length (& end) terminators
-- nb: eu_char 0 is reserved for stop
    integer pad
    sequence packedline
    
    -- pad out to multiple of four
    pad = 4 - and_bits(length(line), 3)
    if pad != 4 then -- ignore exact mulitple of 4
        line &= repeat(0, pad)
    end if

    -- pack 4 eu-chars into 1 eu-integer
    packedline = {}
    for i=1 to length(line) by 4 do
        packedline &= packChars(line[i..i+3])
    end for

    -- tack on line length - including the length integer itself
    packedline = length(packedline)+1 & packedline

    return packedline
end function
-------------------------------------------------------------------------------
global function PackedToAnalysedLine(sequence packedline)
-- takes a line of packed eu_integers and unpacks them into eu_chars
-- nb: length indicator must not be present

    sequence result
    integer len

    result = {}

    for i = 1 to length(packedline) do
        result &= unpackChars(packedline[i])
    end for

    len = length(result) -- search for null terminator

    if len = 0 then -- nil-length line
        return result
    end if

    while 1 do
        if result[len] then
            exit
        end if
        len -= 1
    end while

    return result[1..len]
end function

-------------------------------------------------------------------------------
-------------------------------------------------------------------------------

-- the remaining code analyses the text data and colourizes it to a stage suitable
-- for its storage in the text & colour buffer in screen.e

-------------------------------------------------------------------------------
-------------------------------------------------------------------------------
include colour.e -- global indici etc plus colours

include languages.e -- translation for menus, keywords and builtins
include ini.e
switchToIniGroup("language")
Set_Language( getIniValue( "language", "English" ) )

global sequence win32libs, glocals
win32libs={}
glocals={}

constant -- these map directly to keywords.e
-- used by precedingTabs() function
    IF = 8,
    END = 18,
    PROCEDURE = 7,
    ELSE = 10,
    FOR = 12,
    ELSIF = 11,
    WHILE = 15,
    TYPE = 4,
    FUNCTION = 5


constant
        KEYWORDTOP = length(keywords),
        COMMENT = 30, -- assign a token value to allow quick processing of comments
        TAB = 31, -- remap tab from \t to 31 to allow continuous increments for keywords
        
        ASCIITOP = 127
        
    -- CRITICAL TEST
    ----------------

if hiAsciiTest then
    if KEYWORDTOP>26 -- currently 24!
    or ASCIITOP + length(builtins) >= eu_char then 
    -- overflow has occured not enough space for all builtins
        puts(1, "\n\nOverflow has occured not enough space for all builtins\n\n")
        abort(0)    -- [leave in]
    end if
else
    if ASCIITOP + length(builtins) + 2 >= eu_char then -- this is the max + 1 for eu-chars
        -- overflow has occured not enough space for all builtins
        puts(1, "\n\nOverflow has occured not enough space for all builtins\n\n")
        abort(0)    -- [leave in]
    end if
end if

constant -- character class constants for analyseline(), order is important
    LETTER = 0,
    UNDERSCORE = 1,
    DIGIT = 2,
    QUOTE = 3,
    SYMBOL = 4,
    BRACE = 5,
    OTHER = 6,  
    HIGHASCII=7

-------------------------------------------------------------------------------

function makeCharacterClass()
-- set up character classes for easier line scanning
-- (assume no 0 char)
    sequence symbols, junk
    symbols = ".+-*/^%<>=!#,?[]{}()&"
    if hiAsciiTest then
        junk = repeat(OTHER, 255)
        junk[128..255]=HIGHASCII
    else
        junk = repeat(OTHER, 127)
    end if
    for i = 1 to length(symbols) do
        junk[symbols[i]] = SYMBOL
    end for
    junk['a'..'z'] = LETTER
    junk['A'..'Z'] = LETTER
    junk['_'] = UNDERSCORE
    junk['0'..'9'] = DIGIT
    junk['"' ]  = QUOTE
    junk['\''] = QUOTE
    junk['{'] = BRACE
    junk['}'] = BRACE
    if not singlecolourbraces then
        junk['['] = BRACE
        junk[']'] = BRACE
        junk['('] = BRACE
        junk[')'] = BRACE
    end if
    return junk
end function

constant CLASS = makeCharacterClass()

-------------------------------------------------------------------------------
-- ADD SOME FUNCTIONS TO REMAP THOSE PESKY TABS
constant TABDISPLACEMENT = TAB - '\t'

-- convert all '\t' to TAB
function remapTabs(sequence s)
    sequence hiAsc
    s += (s = '\t') * TABDISPLACEMENT
    if hiAsciiTest then
        hiAsc=s>ASCIITOP
        if find(1,hiAsc) then
            for i=length(hiAsc) to 1 by -1 do
                if hiAsc[i] then
                    s=s[1..i-1]&29&s[i]-ASCIITOP&s[i+1..length(s)]
                end if
            end for
        end if
    end if
    return s
end function

-- convert all TAB's back to '\t', complement of remapTabs()
function unmapTabs(sequence s)
    integer hiAsc
    s -= (s = TAB) * TABDISPLACEMENT
    while 1 do
        hiAsc=find(29,s)
        if not hiAsc then exit end if
        s=s[1..hiAsc-1]&s[hiAsc+1]+ASCIITOP&s[hiAsc+2..length(s)]
    end while
    return s
end function

function unmaphiTop(sequence s)
    integer hiAsc
    while 1 do
        hiAsc=find(29,s)
        if not hiAsc then exit end if
        s=s[1..hiAsc-1]&s[hiAsc+1]+ASCIITOP&s[hiAsc+2..length(s)]
    end while
    return s
end function

        
-------------------------------------------------------------------------------	
-- new option PL 23/09/02 "Replace tabs with spaces when saving file"
-- set precisely once in files.e, saveDiskFile(). 18/11/02 also in context.ew
global integer ExpandTabsToSpaces
ExpandTabsToSpaces=0

global function TranslateLineUserToEuphoria(sequence s)
-- convert translated line
    integer c
    sequence result
 
    integer nextAsHighAscii, line_with_include  
    nextAsHighAscii=0 
    line_with_include=0

    result = {} 
    for i = 1 to length(s) do
        c = s[i]
        if hiAsciiTest and nextAsHighAscii then
            result&=s[i]+ASCIITOP
            nextAsHighAscii=0
        elsif c <= KEYWORDTOP and line_with_include = 0 then -- plug in keyword
            result &= euphoria_keywords[c]
            if equal(euphoria_keywords[c], "include") then line_with_include = 1 end if
        elsif c = TAB then -- found tab
            if ExpandTabsToSpaces then
                c=getTabWidth()
                c=(floor(length(result)/c)+1)*c-length(result)
                result&=repeat(' ',c)
            else
                result &= '\t'
            end if
        elsif c = COMMENT then -- comment found
            result &= "--" & unmapTabs( s[i+1..length(s)] )
            exit
        elsif c=29 then -- high ascii leadin
            nextAsHighAscii=1
        elsif not hiAsciiTest and c>=eu_char-2 then -- a marker for win32lib,glocals
            -- no action; it is a marker character for colours only.
        elsif c > ASCIITOP then -- builtin index found
            c -= ASCIITOP
            result &= euphoria_builtins[c]
        elsif not hiAsciiTest or c>TAB then
            result &= c
        else
            if c <= KEYWORDTOP then result &= keywords[c] end if
            if c > ASCIITOP then result &= builtins[c] end if
        end if
    end for
    return result
end function

global function AnalysedToTextLine(sequence s)
-- convert analysed line into raw text string minus \n
    sequence result
    integer c

    integer nextAsHighAscii
    nextAsHighAscii=0

    result = {}
    for i = 1 to length(s) do
        c = s[i]
        if hiAsciiTest and nextAsHighAscii then
            result&=s[i]+ASCIITOP
            nextAsHighAscii=0
        elsif c <= KEYWORDTOP then -- plug in keyword
            result &= keywords[c]
        elsif c = TAB then -- found tab
            if ExpandTabsToSpaces then
                c=getTabWidth()
                c=(floor(length(result)/c)+1)*c-length(result)
                result&=repeat(' ',c)
            else
                result &= '\t'
            end if
        elsif c = COMMENT then -- comment found
            result &= "--" & unmapTabs( s[i+1..length(s)] )
            exit
        elsif c=29 then -- high ascii leadin
            nextAsHighAscii=1
        elsif not hiAsciiTest and c>=eu_char-2 then -- a marker for win32lib,glocals
            -- no action; it is a marker character for colours only.
        elsif c > ASCIITOP then -- builtin index found
            c -= ASCIITOP
            result &= builtins[c]
        elsif not hiAsciiTest or c>TAB then
            result &= c
        end if
    end for
    return result
end function

global function filterLine(sequence line)
    integer c
    for i = 1 to length(line) do
        c = line[i]
        if hiAsciiTest then
            if c<32 and c!='\t' then
                line[i]=127
            end if
        else
            if not ( (c >= 32 and c <= 127) or c = '\t' ) then -- non ascii value found
                line[i] = 127
            end if
        end if
    end for
    return line
end function

-------------------------------------------------------------------------------

global function TranslateLineEuphoriaToUser(sequence line)
-- analyse text line to an intermediate state
-- suitable for storing (compressing). Newline chars not allowed
-- result is a 1-d eu_char SEQUENCE with indici for keywords & builtins &  comments
-- plus tab is remapped

-- 1-24		index for keywords
-- 25-29	[unused]
-- 30		comment indicator
-- 31		re-mapped TAB
-- 32-127	ascii chars
-- 128-288	index for builtin routines - some excess space is here to absorb future additions
-- 289,291  marker for win32lib, glocals PL 27/10/02

    integer c, start, len, temp, a, b, class, line_with_include
    integer rem, endof, tabcheck
    sequence text, result

    start = 1
    len = length(line)
    result = {}
    tabcheck=1
    line_with_include=0
    
    while start <= len do

        c = line[start]

        start += 1

        -- TABS
        if c = '\t' then
            result &=  TAB
            tabcheck=0
        --  COMMENTS
        elsif c = '-' then -- possible comment
            if start <= len and line[start] = '-' then -- definite comment
                result &= "--" & remapTabs( line[start+1..len] )
                exit
            else -- no comment, only probably negative number found
                result &= c
            end if
        elsif tabcheck and c = ' ' then -- possible tab
            rem=remainder(start-1,getTabWidth())
            endof=start+getTabWidth()-rem-1
            if rem!=0 and endof<=len then
                for i = start to endof do
                    if line[i]!=' ' then
                        result &=c
                        exit
                    end if
                    if i = endof then
                        result&=TAB
                        start=endof+1
                        exit
                    end if
                end for
            else
                result &= c
            end if
        else

            class = CLASS[c]

            --------LETTER/LABEL
            if class = LETTER then
                temp = len
                for i = start to len do
                    c = line[i]
                    class = CLASS[c]
                    if class > DIGIT then
                        temp = i - 1
                        exit
                    end if
                end for
                text = line[start-1..temp]
                start = temp + 1
                if line_with_include = 1 then
                    result &= text
                else
                    c = find(text, euphoria_keywords)
                    if c > 0 then -- keyword found
                        result &= keywords[c]
                        if equal(euphoria_keywords[c], "include") then line_with_include = 1 end if
                    else
                        c = find(text, euphoria_builtins)
                        if c > 0 then -- builtin routine  found
                            result &= builtins[c]
                        else
                            c=find(text,win32libs)
                            if c>0 then -- win32lib word found
                                if hiAsciiTest then
                                    result&=27
                                else
                                    result&=eu_char-2   -- marker byte precedes text
                                end if
                            else
                                c=find(text,glocals)
                                if c>0 then -- local global found
                                    if hiAsciiTest then
                                        result&=28
                                    else
                                        result&=eu_char-1 --marker byte precedes text
                                    end if
                                end if
                            end if
                            result &= text
                        end if
                    end if
                end if

            --------QUOTES 
            elsif class = QUOTE then
                b = len
                a = start - 1
                while start <= len do
                    temp = line[start]
                    if temp = '\\' then -- escape char
                        start += 1
                    elsif temp = c then -- end quote found
                        b = start
                        start += 1
                        exit
                    end if
                    start += 1
                end while
                result &= remapTabs( line[a..b] )

--			--------HIGH ASCII
            elsif class = HIGHASCII then
                result &= 29&c-ASCIITOP

            --------OTHER
            else
                result &= c
            end if

        end if

    end while

    return result

end function

-------------------------------------------------------------------------------
global function TextToAnalysedLine(sequence line)
-- analyse text line to an intermediate state
-- suitable for storing (compressing). Newline chars not allowed
-- result is a 1-d eu_char SEQUENCE with indici for keywords & builtins &  comments
-- plus tab is remapped

-- 1-24		index for keywords
-- 25-29	[unused]
-- 30		comment indicator
-- 31		re-mapped TAB
-- 32-127	ascii chars
-- 128-212	index for builtin routines - some excess space is here to absorb future additions
-- 213,214  marker for win32lib, glocals PL 27/10/02

--DEV: PL 020603: fudge this back:
-- win32lib, glocals held as 27 & 28 & text
--DEVLI, no, that would be highascii:
-- builtins held as 29 & code
-- change the 127 limit in editor.exw
-- and see changes marked 020603 above and below...
-- ..or use a lead-in meaning the next char is ascii 128..255
--
    integer c, start, len, temp, a, b, class
    integer rem, endof, tabcheck
    sequence text, result

    start = 1
    len = length(line)
    result = {}
    tabcheck=1
    
    while start <= len do

        c = line[start]

        start += 1

        -- TABS
        if c = '\t' then
            result &=  TAB
            tabcheck=0
        --  COMMENTS
        elsif c = '-' then -- possible comment
            if start <= len and line[start] = '-' then -- definite comment
                result &= COMMENT & remapTabs( line[start+1..len] )
                exit
            else -- no comment, only probably negative number found
                result &= c
            end if

        elsif tabcheck and c = ' ' then -- possible tab
            rem=remainder(start-1,getTabWidth())
            endof=start+getTabWidth()-rem-1
            if rem!=0 and endof<=len then
                for i = start to endof do
                    if line[i]!=' ' then
                        result &=c
                        exit
                    end if
                    if i = endof then
                        result&=TAB
                        start=endof+1
                        exit
                    end if
                end for
            else
                result &= c
            end if
        else

            class = CLASS[c]

            --------LETTER/LABEL
            if class = LETTER then
                temp = len
                for i = start to len do
                    c = line[i]
                    class = CLASS[c]
                    if class > DIGIT then
                        temp = i - 1
                        exit
                    end if
                end for
                text = line[start-1..temp]
                start = temp + 1
                c = find(lower(text), lower(keywords))
                if c > 0 then -- keyword found
                    result &= c
                else
                    c = find(lower(text), lower(builtins))
                    if c > 0 then -- builtin routine  found
                        result &= c + ASCIITOP
                    else
                        c=find(text,win32libs)
                        if c>0 then -- win32lib word found
                            if hiAsciiTest then
                                result&=27
                            else
                                result&=eu_char-2   -- marker byte precedes text
                            end if
                        else
                            c=find(text,glocals)
                            if c>0 then -- local global found
                                if hiAsciiTest then
                                    result&=28
                                else
                                    result&=eu_char-1 --marker byte precedes text
                                end if
                            end if
                        end if
                        result &= text
                    end if
                end if

            --------QUOTES 
            elsif class = QUOTE then
                b = len
                a = start - 1
                while start <= len do
                    temp = line[start]
                    if temp = '\\' then -- escape char
                        start += 1
                    elsif temp = c then -- end quote found
                        b = start
                        start += 1
                        exit
                    end if
                    start += 1
                end while
                result &= remapTabs( line[a..b] )

--			--------HIGH ASCII
            elsif class = HIGHASCII then
                result &= 29&c-ASCIITOP

            --------OTHER
            else
                result &= c
            end if

        end if

    end while

    return result

end function


-------------------------------------------------------------------------------

global function makeBufferLine(sequence line, integer tabwidth)
-- using an analysed line construct a dual (text & colour) line
-- and efficiently assemble it for inclusion in the buffer

    sequence sText, sColour, temp2, texttemp
    integer c, class, len, temp, start, len_sText, colour
    object junk, text

    len = length(line)

    -- early out?
    if len = 0 then -- take care of nil-length lines
        return { {}, {} }
    end if

    -- initialize vars
    sText = {}
    sColour = {}
    colour = COLOURSPACE
    texttemp = {}
    start = 1
   

    while 1 do

        c = line[start]

        if c <= KEYWORDTOP then -- plug in keyword
            text = keywords[c]
            c = COLOURKEYWORD

        elsif c = COMMENT then
            if hiAsciiTest then
                text = "--" & unmaphiTop(line[start+1..len])
            else
                text = "--" & line[start+1..len]
            end if
            c = COLOURCOMMENT
            start = len

        elsif c = TAB then
            text = repeat(' ', tabwidth - remainder(length(sText)+ length(texttemp), tabwidth))
            c = COLOURSPACE
            if isChecked(ToolStab) then
                c=COLOURHILITE
            end if
        elsif c = ' ' then
            text = " "
            c = COLOURSPACE
            if isChecked(ToolStab) then
                c=COLOUROTHER
            end if
        elsif (hiAsciiTest and find(c,{27,28})) or (not hiAsciiTest and c>=eu_char-2) then -- marker for win32lib,glocals
            temp=start+1
            for i=temp to len do
                if CLASS[line[i]]>DIGIT then -- all done
                    exit
                end if
                temp=i
            end for         
            text=line[start+1..temp]
            start=temp

            if (hiAsciiTest and c=27)   or (not hiAsciiTest and c=eu_char-2) then
                c=COLOURWIN32LIB
            else
                c=COLOURGLOCALS
            end if
        elsif hiAsciiTest and c=29 then
            start+=1
            text={line[start]+ASCIITOP}
            c=COLOUROTHER

        elsif c > ASCIITOP then -- plug in builtin
            c -= ASCIITOP
            text = builtins[c]
            c = COLOURBUILTIN

        else -- rest of elements are in ascii range

            class = CLASS[c]

            if class = QUOTE then
                temp = start+1
                while 1 do
                    if temp > len then -- exit if line length exceeded
                        temp = len
                        exit
                    end if
                    junk = line[temp]
                    if junk = '\\' then -- escape char found
                        temp += 1
                    elsif junk = c then -- end quote of same type found
                        exit
                    elsif junk = TAB then -- forget string et al, show that tab!
                        --
                        -- the string colour will be messed up, remainder of string will
                        -- not be in string colour and likely remainder of line after
                        -- the closing quote will likely be incorrectly string colour.
                        -- 
                        if isChecked(ToolStab) then temp-=1 exit end if
                    end if
                    temp += 1
                end while
                if hiAsciiTest then
                    text = unmaphiTop(line[start..temp])
                else
                    text = line[start..temp]
                end if
                start = temp
                c = COLOURSTRING

            elsif class = SYMBOL then
                text = {c}
                c = COLOURSYMBOL

            elsif class = BRACE then
                if singlecolourbraces then
                    text = {c}
                    c = COLOURBRACE
                end if
            elsif class = LETTER then
                temp = len
                for i = start + 1 to len do
                    if CLASS[line[i]] > DIGIT then -- non-label character found
                        temp = i - 1
                        exit
                    end if
                end for
                text = line[start..temp]
                start = temp
                c = COLOUROTHER

            else -- OTHER class element detected
                text = {c}
                c = COLOUROTHER
            end if

        end if

        if c >= COLOURSTRING then -- calc. tab spacings for comments (or strings)
            temp2 = {}
            len_sText = length(sText) + length(texttemp)
            for i = 1 to length(text) do
                junk = text[i]
                if junk = TAB then
                    junk = tabwidth - remainder(len_sText + length(temp2), tabwidth)
                    temp2 &= repeat(' ', junk)
                else
                    temp2 &= junk
                end if
            end for
            text = temp2
        end if

        -- make fully efficient coloured line
        if c = COLOURSPACE or c = colour then -- must be at start of line
            texttemp &= text
        elsif colour = COLOURSPACE then -- first coloured object encountered
            colour = c
            texttemp &= text
        elsif c = colour then -- compatible colours
            texttemp &= text
        else
            sColour &= repeat(colour, length(texttemp))
            sText &= texttemp
            texttemp = text
            colour = c
        end if

        start += 1
        if start > len then exit end if

    end while

    -- deal to entire line being whitespace
    if colour = COLOURSPACE then -- whole line whitespace
        colour = COLOUROTHER
    end if

    -- attach last text & colour pieces
    sColour &= repeat(colour, length(texttemp))
    sText &= texttemp

    return {sText, sColour}

end function
-------------------------------------------------------------------------------

global function jumpLeftToWord(sequence text, integer pos)
-- return with the start char of the word/label right of pos etc..
--DEV integer ctp

    -- already against right column?
    if pos <= 1 then
        return 1
    end if

    -- find the next char left of pos
    pos -= 1
    if CLASS[text[pos]] > DIGIT then -- not already in word
        for i = pos to 1 by -1 do
            if i = 1 then
                return 1
            elsif CLASS[text[i]] <= DIGIT then
                pos = i
                exit
            end if
        end for
    end if

    --find very next non-letter char left of pos
    for i = pos to 1 by -1 do
        if CLASS[text[i]] > DIGIT then
            return i+1
        end if
    end for

    return 1

end function


global function jumpRightToWord( sequence text, integer pos )
    integer len

    len = length(text)

    -- already against right column?
    if pos > len then
        return pos
    end if

    -- find the next non-letter char right of pos
    for i = pos to len do
        if CLASS[text[i]] > DIGIT then -- not already in word
            pos = i
            exit
        elsif i = len then
            return len + 1
        end if
    end for

    -- find very next letter char right of pos
    for i = pos to len do
        if CLASS[text[i]] <= DIGIT then
            return i
        end if
    end for

    return len + 1

end function
    
global function precedingTabs(sequence rawtext, integer cursorcol)
-- get the number of tabs for the succeeding line
-- so that ENTER key will move the cursor across properly to initial position
-- also returns w32True/w32False if line is ONLY made of TABS
    sequence line
    integer tab_count, pos_keyword, r_flag, isMereTabs

    if cursorcol = 1 then
        return {"", 0}
    end if

    line = TextToAnalysedLine( rawtext )
    tab_count = 0
    r_flag = 0
    isMereTabs = w32False

    -- count number of leading TABS in the line
    for i = 1 to length(line) do
        if line[i] != TAB then
            exit
        end if
        tab_count += 1
    end for

    -- determine if line is entirely tabs
    if tab_count and tab_count = length(line) then
        isMereTabs = w32True
    end if

    --	deal to special cases
    pos_keyword = find(IF, line)
    if not pos_keyword then
        pos_keyword = find(ELSIF, line)
        if not pos_keyword then
            pos_keyword = find(ELSE, line)
            if not pos_keyword then
                pos_keyword = find(PROCEDURE, line)
                if not pos_keyword then
                    pos_keyword = find(FOR, line)
                    if not pos_keyword then
                        pos_keyword = find(WHILE, line)
                        if not pos_keyword then
                            pos_keyword = find(FUNCTION, line)
                            if not pos_keyword then
                                pos_keyword = find(TYPE, line)
                                if pos_keyword then
                                    tab_count=0 pos_keyword=0
                                end if
                            else
                                tab_count=0 pos_keyword=0
                            end if
                        end if
                    end if
                else
                    tab_count=0 pos_keyword=0
                end if
            end if
        end if
    end if

    if not find(END, line) then
        if pos_keyword then
            tab_count += 1
        end if
    elsif r_flag then
        tab_count = 0
    end if
    
    return {repeat( '\t', tab_count), isMereTabs}
end function