without trace

-------------------------------------------------------------------------------
-- This section of code deals with the editing of lines
-- and holds the edit params
-------------------------------------------------------------------------------


integer cursorline, cursorcolumn
    cursorline = 1
    cursorcolumn = 1

global function getCursorLine()
    return cursorline
end function

global function getCursorColumn()
    return cursorcolumn
end function


--------------------------------------------------------------------------------
procedure storeEditLine()
    replaceFileLine( getEditLine(), getEditLineNumber() )
end procedure

global procedure moveToNewEditLine(integer newline)
    if getEditLineNumber() != newline then
        storeEditLine()
        setEditLine(newline)
    end if
end procedure

global procedure storecurrentstate()
    if not getNumberOfFiles() then
        return
    end if
    setVal(TOPLINE, getTopLine() )
    setVal(LEFTCOLUMN, getLeftColumn() )
    setVal(CURSORLINE, cursorline )
    setVal(CURSORCOLUMN, cursorcolumn )
    setVal(SELECTION, getAnchorDragAll() )
    storeEditLine()
end procedure

global procedure retrievecurrentstate()
    if not getNumberOfFiles() then
        return
    end if
    setTopLine ( getVal(TOPLINE) )
    setLeftColumn ( getVal(LEFTCOLUMN) )
    cursorline = getVal(CURSORLINE)
    cursorcolumn = getVal(CURSORCOLUMN)
    setAnchorDragAll( getVal(SELECTION) )
    -- reset editline
    setEditLineNumber(0) -- force file line access
    setEditLine(getVal(CURSORLINE))
end procedure

-------------------------------------------------------------------------------
--------------------------------cursor area------------------------------------
-------------------------------------------------------------------------------


global function isCursorOnScreen() -- test to see if the cursor coords should be visible
    if cursorline >= getTopLine() and cursorline < getTopLine() + getScreenRows() and
        cursorcolumn >= getLeftColumn() and cursorcolumn < getLeftColumn() + getScreenColumns() then
        return w32True
    end if
    return w32False
end function

global procedure setCursorPosition(integer line, integer col)
 -- calculate cursor coords to their screen position
    integer x, y

    -- set cursor line coords
    cursorline = line
    cursorcolumn = col

    -- clip to make relative to screen
    x = cursorcolumn - getLeftColumn() + 1
    y = cursorline - getTopLine() + 1

    -- expand to start pixel coords of the MAIN window
    --x = x * getFontWidth() - getFontWidth() + editx
    --y = y * getFontHeight() - getFontHeight() + edity

    -- set cursor pixel coords
    setCursorAt(x, y)

end procedure


global procedure hideCursor()
    hideMyCursor()
end procedure

global procedure showCursor()
    if getNumberOfFiles() and isCursorOnScreen() and isFileTabVisible( getCurrentFileNum() ) then
        setCursorPosition(cursorline, cursorcolumn)
        showMyCursor()
    else
        hideCursor()
    end if
end procedure


global procedure moveCursorLeft()
    integer line, col

    line = cursorline
    col = cursorcolumn

    if col = 1 then -- cursor hard against left boundary
        if line > 1 then -- and not on very first line, so..
            line -= 1
            storeEditLine()
            setEditLine( line )
            col = LinePrintLength( getEditLine() ) + 1
        else
            -- first line & first column, do nothing
        end if
    else -- only have to move left along same line
        col = PrintCharStartPos(col - 1, getEditLine())
    end if

    if not (line = cursorline and col = cursorcolumn) then -- movement has occurred
        setCursorPosition(line, col)
    end if

end procedure


global procedure moveCursorRight()
    integer line, col

    line = cursorline
    col = cursorcolumn

    if col = LinePrintLength( getEditLine() ) + 1 then -- col on very right side of line
        if line < getNumberOfLinesInFile() then -- ok, can move down
            line += 1
            storeEditLine()
            setEditLine( line )
            col = 1
        else
            -- cursor at end of file, do nothing
        end if
    else -- move forward one space
        col = PrintCharEndPos( col, getEditLine() ) + 1
    end if

    if not (line = cursorline and col = cursorcolumn) then -- movement has occurred
        setCursorPosition(line, col)
    end if

end procedure


global procedure moveCursorUp()
    integer line, col

    line = cursorline
    col = cursorcolumn

    if line > 1 then
        line -= 1
        storeEditLine()
        setEditLine( line )
        col = PrintCharStartPos( col, getEditLine() )
        setCursorPosition(line, col)
    end if

end procedure


global procedure moveCursorDown()
    integer line, col

    line = cursorline
    col = cursorcolumn

    if line < getNumberOfLinesInFile() then
        line += 1
        storeEditLine()
        setEditLine( line )
        col = PrintCharStartPos( col, getEditLine() )
        setCursorPosition(line, col)
    end if

end procedure


global procedure jumpCursorLeft()
    integer col
    sequence text

    col = getCursorColumn()


    -- get the line of text that cursor is on
    text = getTextLine( getCursorLine() )

    -- determine new position
    col = PrintToBytePos(col, text)

    --if cursor at start of line then will jump to next line up
    if col = 1 then -- try to move up to next line
        moveCursorLeft()
        return
    end if

    -- calculate position of next word
    col = jumpLeftToWord( text,  col )
    col = ByteToPrintPos(col, text)

    setCursorPosition(getCursorLine(), col)

end procedure


global procedure jumpCursorRight()
    integer col
    sequence text

    col = getCursorColumn()

    -- get the line of text that cursor is on
    text = getTextLine( getCursorLine() )

    -- determine new position
    col = PrintToBytePos(col, text)

    -- if cursor at end of line then will jump to next line down
    if col > length(text) then -- try to move down to next line
        moveCursorRight()
        return
    end if

    -- calculate position of next word
    col = jumpRightToWord( text,  col )
    col = ByteToPrintPos(col, text)

    --if col != getCursorColumn() then -- movement has occurred
        setCursorPosition(getCursorLine(), col)
    --end if

end procedure
-------------------------------------------------------------------------------
-------------------------clipboard functions-----------------------------------
-------------------------------------------------------------------------------


global procedure placetextonclipboard(sequence text)
-- convert to 1-d sequence with adj for c/r
-- put on board
    sequence s

    s = text[1]

    for i = 2 to length(text) do
        --This *should be quite a bit faster...
        s&="\r\n"
        s&=text[i]
    end for

    -- attempt to place text on clipboard
    if copyTextToClipboard(MAIN, s) then
        -- success
    end if
    setEnable(EditPste, w32True)

end procedure


global function gettextfromclipboard()
    object cbtext
    sequence text
    integer len, start, p, c

    -- try to retrieve text from clip board
    cbtext = getTextFromClipboard(MAIN)

    if integer(cbtext) or not length(cbtext) then -- nothing to get
        return {} -- presume nothing to acquire
    end if

    -- remove char 250 apparently at end of text sometimes
    len = length(cbtext)
    if cbtext[len] = 250 then
        cbtext = cbtext[1 .. len]
    end if

    -- use any "\r\n" combinations to determine where the newline will occur
    -- also replace any non-ascii values with char 127
    len = length(cbtext)
    start = 1
    p = 1
    text = {}

    while 1 do
        c = cbtext[p]
        if c = '\r' then
            text = append(text, cbtext[start .. p - 1] )
            start = p + 2
            p = start
            if p > len then
                exit
            end if
        else
            if (hiAsciiTest=0 and c > 127) or (c < ' ' and not(c = '\t' or c = '\n')) then -- convert values above 127
                cbtext[p] = 127
            end if
            p += 1
            if p > len then
                text = append(text, cbtext[start .. p - 1] )
                exit
            end if
        end if
    end while
        
    return text
end function



-------------------------------------------------------------------------------
-------------------------------get text area-----------------------------------
-------------------------------------------------------------------------------


global function gettext(integer top, integer left, integer bottom, integer right)

-- retrieve the text from the ref values

    sequence line, textlines

    line = getTextLine(top)

    -- deal to single line
    if top = bottom then
        left = PrintToBytePos(left, line)
        right = PrintToBytePos(right, line)
        return { line[left .. right] }
    end if

    -- else, grab part of first (top) line
    left = PrintToBytePos( left, line )
    textlines = { line[ left .. length( line ) ] }

    -- if there are any lines in the middle of top & bottom then add these to the return var
    for i = top + 1 to bottom - 1 do -- this will only execute for valid line values
        textlines = append(textlines, getTextLine(i))
    end for

    -- deal to last line
    line = getTextLine(bottom)
    left = 1
    right = PrintToBytePos(right, line)
    textlines = append( textlines, line[left .. right] )

    return textlines
end function


global function getTextSelection()
    if isTextSelected() then
        return gettext ( getSelectTop(), getSelectLeft(), getSelectBottom(), getSelectRight() )
    else
        return {}
    end if
end function



-------------------------------------------------------------------------------
-----------------------------delete text area----------------------------------
-------------------------------------------------------------------------------


function deletetext(integer top, integer left, integer bottom, integer right)

    integer left_, right_
    sequence lastline, textlines, editline

    -- store editline if not on same line as 'top' line
    -- top line will be the resultant line afterwards
    moveToNewEditLine( top )

    -- remove line lengths of all lines affected
    for i = top to bottom do
        removeLineLength( getTextLine( i ) )
    end for

    -- prepare scratch vars
    editline = getEditLine()

    -- deal to single line
    if top = bottom then
        left_ = PrintToBytePos(left, editline)
        right_ = PrintToBytePos(right, editline)
        textlines = { editline[ left_ .. right_ ] }
        setEditLine( editline[1 .. left_ - 1] & editline[right_ + 1 .. length(editline)] )
        addLineLength( getEditLine() )
        setCursorPosition(top, left)
        setScreenLines( top, top )
        setDisplayRange ( top, top )
        return { top, left, bottom, right, textlines } 
    end if

    -- get pre-select text from first line & initialize the return variable
    left_ = PrintToBytePos( left, editline )
    textlines = { editline[ left_ .. length( editline ) ] }
    editline = editline[1 .. left_ - 1]

    -- if there are any lines in the middle of top & bottom then add these to the return var
    for i = top + 1 to bottom - 1 do -- this will only execute for valid line values
        lastline = getTextLine(i)
        textlines = append(textlines, lastline)
    end for

    -- get post-select text from last line and make fresh line & add last piece to return var
    lastline = getTextLine(bottom)
    right_ = PrintToBytePos(right, lastline)
    textlines = append( textlines, lastline[ 1 .. right_ ] )
    editline &= lastline[ right_ + 1 .. length(lastline) ]

    -- add back the new line length
    addLineLength(editline)

    -- re-appoint the editline
    setEditLine(editline)

    -- re-print the edit line in the buffers
    setScreenLines(top, top)    

    -- delete all lines BELOW the first line (which will become the cursorline)
    deleteFileLines(top + 1, bottom)
    deleteScreenLines(top + 1, bottom)

    -- update the cursor positions
    setCursorPosition(top, left)

    -- prepare display values
    setDisplayRange (top, getTopLine() + getScreenRows() - 1 )

    return { top, left, bottom, right, textlines }
end function


-------------------------------------------------------------------------------
-----------------------------insert text area----------------------------------
-------------------------------------------------------------------------------


    
function inserttext(integer top, integer left, sequence textlines, integer hFlag)

    integer bottom, right, pos
    sequence first, last, textblock, line

    bottom = top + length(textlines) - 1

    -- store editline if not on same line as 'bottom' line (where cursor will be)
    moveToNewEditLine( top )

    line = getEditLine()

    -- remove line length for the line in which insertion will be made
    removeLineLength( line )

    -- chop up the 'insertee' line
    pos = PrintToBytePos(left, line)
    first = line[1 .. pos - 1] --	first is START of new text block
    last = line[pos .. length(line)] -- last is END of new text block

    -- process and exit if single line only
    if length(textlines) = 1 then
        line = first & textlines[1]
        right = LinePrintLength( line )
        setCursorPosition( bottom, right + 1 )
        line &= last
        addLineLength( line )
        setEditLine( line )
        if hFlag then
            setSelectAll(top, left, bottom, right)
        end if
        setScreenLines( top, top )
        setDisplayRange ( top, top )
        return {top, left, bottom, right, textlines}
    end if

    -- create insertable text block and add in the new line lengths for all but last line
    textblock = { first & textlines[1] }
    addLineLength( textblock[1] )
    for i = 2 to length(textlines) - 1 do
        addLineLength( textlines[i] )
        textblock = append( textblock, textlines[i] )
    end for

    -- extract text from last line to be inserted
    first = textlines[length(textlines)]

    -- prepare a return coord
    right = LinePrintLength( first )

    -- construct & update editline
    setEditLine ( first & last )
    setEditLineNumber(bottom)

    -- add in final line's length
    addLineLength( getEditLine() )

    -- insert file lines
    insertFileLines(textblock, top )

    -- set highlight params if needed
    if hFlag then
        setSelectAll(top, left, bottom, right)
    end if

    -- update lines in buffers (t&c and pixmap)
    insertScreenLines(top, bottom)

    -- update edit params
    setCursorPosition(bottom, right + 1)

    -- update display params
    setDisplayRange ( top, getTopLine() + getScreenRows() - 1 )

    return {top, left, bottom, right, textlines}

end function




-------------------------------------------------------------------------------

constant

    -- action structure
    MSG = 1,
    PARAMS = 2,
    DATA = 3,

    -- action messages
    INSERTTEXT = 1,
    DELETETEXT = 0

-------------------------------------------------------------------------------

function performAction(integer msg, sequence params, sequence date, integer hFlag)
-- this routine will interpret & execute the action only

    object result -- this will hold the line/s of characters to be stored

    if msg = INSERTTEXT then
        result = inserttext(params[1],params[2], date, hFlag)
    elsif msg = DELETETEXT then
        result = deletetext(params[1], params[2], params[3], params[4])
    end if

    -- set flag requiring file to be saved after edition
    if getVal(SAVEFLAG)=w32False then   --PL 23/03/02 maintain "*" flag
        setVal( SAVEFLAG, w32True )
        assertSelectTab()
        setfilenameincaptionbar()
        transferTabbarToWindow(MAIN)
    end if
    
    return result
end function

--------------------------------------------------------------------------------------

global procedure reverseAction() -- undo last action
    integer msg, hFlag
    sequence params, date
    object action

    action = getPreviousAction()

    -- was there a prev action?
    if integer(action) then
        return
    end if

    -- Get last action
    msg = action [MSG]
    params = action [PARAMS]
    date = action [DATA]

    -- Reverse event messages
    if msg = INSERTTEXT then
        msg = DELETETEXT
        hFlag = w32False
    elsif msg = DELETETEXT then
        msg = INSERTTEXT
        hFlag = w32True
    end if

    -- execute event
    date = performAction(msg, params, date, hFlag)

end procedure 

--------------------------------------------------------------------------------------

global procedure forwardAction() -- redo past actions
    integer msg, hFlag
    sequence params, date
    object action

    action = getNextAction()

    -- is there a next action?
    if integer(action) then
        return
    end if

    -- Get last action
    msg = action [MSG]
    params = action [PARAMS]
    date = action [DATA]

    -- ensure highlighting done on insertion
    if msg = INSERTTEXT then
        hFlag = w32True
    elsif msg = DELETETEXT then
        hFlag = w32False
    end if

    -- execute event
    date = performAction(msg, params, date, hFlag)

end procedure 

--------------------------------------------------------------------------------------

procedure processAction(integer msg, sequence params, sequence date, integer hFlag)
    sequence junk

    junk = performAction(msg, params, date, hFlag)

    addNewAction(msg, junk[1..4], junk[5])

end procedure

-------------------------------------------------------------------------------
-------------------------------text event area---------------------------------
-------------------------------------------------------------------------------

global procedure deleteSelection()
    integer top, left, bottom, right
    top = getSelectTop()
    left = getSelectLeft()
    bottom = getSelectBottom()
    right = getSelectRight()
    resetSelection()
    processAction ( DELETETEXT, { top, left, bottom, right }, { }, w32False )
end procedure

global procedure inSertText(sequence text)
    processAction ( INSERTTEXT, { cursorline, cursorcolumn }, text, w32False )
end procedure

global procedure insertHighlitText(sequence text)
    processAction ( INSERTTEXT, { cursorline, cursorcolumn }, text, w32True )
end procedure

global procedure deleteText()

    -- test for cursor at end of file
    if cursorline = getNumberOfLinesInFile() and cursorcolumn > LinePrintLength( getTextLine(cursorline) ) then
        return --  cannot delete when at end of file
    -- test for cursor at end of line
    elsif cursorcolumn > LinePrintLength( getTextLine(cursorline) ) then
        processAction ( DELETETEXT, { cursorline, cursorcolumn, cursorline + 1, 0 }, {}, w32False )
    -- normal delete action
    else
        processAction ( DELETETEXT, { cursorline, cursorcolumn, cursorline, cursorcolumn }, {}, w32False )
    end if
    
end procedure