-- this code will manage the files opened in the editor and state variables of those files
include get.E
include file.E
--------------------
global function convertToLowerCase(sequence text)
    integer disp, c
    disp = 'a' - 'A'
    for i = 1 to length(text) do
        c = text[i]
        if c <= 'Z' and c >= 'A' then
            text[i] = c + disp
        end if
    end for
    return text
end function
--------------------

sequence editline
integer editlinenumber

editline = ""
editlinenumber = 0 -- force file line access because 2nd test in setEditLine() will fail

sequence status -- 2d sequence to hold state of all opened files & used to restore prev session
sequence files, -- 2d sequence containing all opened files
         linelengths, -- 2d sequence holding the line lengths of all opened files
         currentindex -- holds index value of each line in the current(uppermost) file

integer currentfilepos -- position in queue for uppermost opened file


global procedure resetFileVariables()
    files = {} -- initialize
    status = {}
    linelengths = {}
    currentindex = {1}
    currentfilepos = 0
    updateMainMenu(w32False)    --disable
end procedure

resetFileVariables()

global function getCurrentFileNum()
    return currentfilepos
end function

global function getStatusValue(integer filepos, integer variable)
    return status[filepos][variable]
end function

global function getNumberOfFiles() -- get number of files currently opened
    return length(files)
end function

global constant -- these are to access state variables in 'status'
    FILEPATH = 1,
    FILENAME = 2,
    SAVEFLAG = 3,

    TOPLINE = 4,
    LEFTCOLUMN = 5,

    CURSORLINE = 6,
    CURSORCOLUMN = 7,

    SELECTION = 8


global procedure setVal(integer variable, object value) -- update *current* file state variable
    status[currentfilepos][variable] = value
end procedure

global function getVal(integer variable) -- retrieve *current* file state variable
    if currentfilepos = 0 then
        if variable <= FILENAME then
            return ""
        else
            ? 9/0 -- [leave in] something wrong
            return 0
        end if
    end if
    return status[currentfilepos][variable]
end function

global function getFileName(integer filepos)
    if filepos = 0 then
        return ""
    else
        return status[filepos][FILENAME]
    end if
end function

global function getFilePath(integer filepos)
    if filepos = 0 then
        return ""
    else
        return status[filepos][FILEPATH]
    end if
end function


include action.E
-----------------------------------------------------------------------------
-------------------------------line lengths----------------------------------
-----------------------------------------------------------------------------



global function getMaxLineLength()
    if currentfilepos then
        return length( linelengths[currentfilepos] ) - 1
    end if
    return 0
end function


global procedure addLineLength(sequence rawtext)
    integer len, disp

    -- get length of new line
    len = LinePrintLength(rawtext)
    len += 1 -- to account for any 0 length lines crashing euphorias 1-based system

    -- expand lengths structure if needed
    if len > length(linelengths[currentfilepos]) then
        disp = len - length(linelengths[currentfilepos])
        linelengths[currentfilepos] &= repeat(0, disp)
    end if

    -- increment value of line at this point
    linelengths[currentfilepos][len] += 1
end procedure

-------------------------------------------------------------------------------

global procedure removeLineLength(sequence rawtext)
    integer len

    len = LinePrintLength(rawtext)
    len += 1 -- to account for any 0 length lines crashing euphorias 1-based system

    if len < length(linelengths[currentfilepos]) then
        linelengths[currentfilepos][len] -= 1
        return -- no need to check for nil entries on end of list

    elsif len > length(linelengths[currentfilepos]) then
        puts(1,"files.e : error in the linelengths system\n\n")
        ?9/0 -- [leave in] major error has occurred

    else
        linelengths[currentfilepos][len] -= 1
        len = 1
        for i = length(linelengths[currentfilepos]) to 1 by -1 do
            if linelengths[currentfilepos][i] then len = i exit end if
        end for

        -- remove excess (nil) end lengths
        if len = length(linelengths[currentfilepos]) then -- no change to max line length
            return
        else
            linelengths[currentfilepos] = linelengths[currentfilepos][1..len]
        end if

    end if
end procedure

-------------------------------------------------------------------------------

function makeNewStatusStructure()
    return {
        {}, -- name
        {}, -- path
        w32True,    -- saveflag - file requires saving 0=no, 1=yes

        1,  -- topline - line at top of screen
        1,  -- leftcolumn - column at left of screen

        1,  -- cursorline - line being edited
        1,  -- cursorcolumn - column where cursor is

        {0, 0, 0, 0} -- selection
    }
end function

--------------------------------------------------------------------------------	

--PL 23/09/02 mothballed
--function ensureFileExtn(sequence newname, sequence oldname)
--	sequence new, old
--
--	new = extractNameAndExtn(newname) -- in  disk.e
--	new = new[2]
--
--	old = extractNameAndExtn(oldname) -- in  disk.e
--	old = old[2]
--
--	if equal(new, old) then
--		return newname
--	elsif not length(new) then
--		return newname & '.' & old
--	end if
--	return newname
--	end function
--

--------------------------------------------------------------------------------


function makeFileIndex(sequence datafile) -- make fresh index for specified file
    sequence result
    integer len, c

    len = length(datafile)
    result = {1} -- first line will always start at 1 and nil sequence will have this too
    c = 1

    if len = 0 then
        return {1,2}
    end if

    while 1 do

        --get length integer and accumulate
        c += datafile[c]

        --add to index list
        result &= c

        --test if end of line reached
        if c > len then
            if c = len + 1 then -- boundary correctly landed
                exit
            else -- error has occurred, landing is too far over limit. should only be len+1
                puts(1, "Error : landing is too far over limit. should only be len+1")
                ? 9/0 -- [leave in]
            end if
        end if

    end while

    return result
end function
    
--------------------------------------------------------------------------------

global function getNumberOfLinesInFile() -- get length of current file
    return length(currentindex) - 1
end function    

--------------------------------------------------------------------------------

global function getPackedLine(integer line) -- return the lengthed & packed line
-- this routine is specifically for routines.e
-- PL 23/03/02 see also compare.ew
    integer start, finish

    if line = editlinenumber then
        return AnalysedToPackedLine(TextToAnalysedLine( editline ))
    end if

    -- calculate boundaries of line desired
    start = currentindex[line]
    finish = currentindex[line + 1] - 1

    -- retrieve packed line data
    return files[currentfilepos][start..finish]

end function
--------------------------------------------------------------------------------

global function getAnalysedLine(integer line)--see gettextline() below
-- will return the ANALYSED line pointed to in the current data-file in analysed form

    integer start, finish
    sequence s

    if line = editlinenumber then
        return TextToAnalysedLine( editline )
    end if

    -- calculate boundaries of line desired
    start = currentindex[line] + 1
    finish = currentindex[line + 1] - 1

    -- retrieve line data & process to analysed state
    s = files[currentfilepos][start..finish]
    s = PackedToAnalysedLine(s)

    return s
end function

--------------------------------------------------------------------------------

global function getTextLine(integer line)--see getAnalysedLine() above
-- will return the TEXT line pointed to in the current file

    sequence textline

    -- line is actually editline?
    if line = editlinenumber then
        return editline
    end if

    -- retrieve line in analysed form
    textline = getAnalysedLine(line)

    -- convert to raw text
    textline = AnalysedToTextLine(textline)

    return textline
end function
--------------------------------------------------------------------------------
--------------------------edit line access--------------------------------------
--------------------------------------------------------------------------------
global procedure setEditLine(object x)
    if integer(x) then -- access new line in current file
        if x != editlinenumber then
            editline = getTextLine(x)
            editlinenumber = x
        end if
    else -- direct assignment
        editline = x
    end if
end procedure

global function getEditLine()
    return editline
end function

global procedure setEditLineNumber(integer n)
    editlinenumber = n
end procedure

global function getEditLineNumber()
    return editlinenumber
end function

--------------------------------------------------------------------------------

global procedure deleteFileLines(integer top, integer bottom)
-- delete lines top to bottom in currentfile. To delete same line do: deleteFileLines(x,x)
    integer start, finish, disp, len

    -- calculate delete lines params
    start = currentindex[top]
    finish = currentindex[bottom + 1]
    disp = finish - start
    len = length(files[currentfilepos])

    -- work out minimum movement required
    if start < len - finish then -- move front of sequence
        files[currentfilepos][disp + 1 .. finish - 1] = files[currentfilepos][1 .. start - 1]
        files[currentfilepos] = files[currentfilepos][disp + 1 .. len]
    else -- move end of sequence
        files[currentfilepos][start .. len - disp] = files[currentfilepos][start + disp .. len]
        files[currentfilepos] = files[currentfilepos][1 .. len - disp]
    end if

    -- update index
    len = length(currentindex)
    currentindex[bottom + 1 .. len] -= disp
    disp = bottom - top + 1
    currentindex[top .. len - disp] = currentindex[bottom + 1 .. len]
    currentindex = currentindex[1 .. len - disp]

end procedure



--------------------------------------------------------------------------------

global procedure insertFileLines(sequence textblock, integer linepos)
    sequence text, line, index
    integer lenfile, lenline, pos

    -- prepare text data for insertion and create the index
    line = {}
    index = {}
    lenline = 0
    for i = 1 to length(textblock) do
        index &= lenline
        text = textblock[i]
        text = TextToAnalysedLine(text) -- make analysed line
        text = AnalysedToPackedLine(text) -- compact into 'lengthed' packed line
        line &= text -- append to line that will eventually be inserted
        lenline += length(text)
    end for

    -- insert space for multi-line sequence into the file
    pos = currentindex[linepos]
    lenfile = length(files[currentfilepos]) + lenline
    if linepos < lenfile - linepos then -- move front of sequence
        files[currentfilepos] = repeat(0, lenline) & files[currentfilepos]
        files[currentfilepos][1 .. pos - 1] = files[currentfilepos][lenline + 1 .. pos + lenline - 1]
    else -- move back of sequence
        files[currentfilepos] &= repeat(0, lenline)
        files[currentfilepos][pos + lenline .. lenfile] = files[currentfilepos][pos .. lenfile - lenline]
    end if

    -- insert the actual multi-line itself
    files[currentfilepos][pos .. pos + lenline - 1] = line

    -- update the line index
    index += currentindex[linepos]
    currentindex[linepos .. length(currentindex)] += lenline
    lenline = length(index)
    currentindex &= repeat(0, lenline)
    lenfile = length(currentindex)
    currentindex[linepos + lenline .. lenfile] = currentindex[linepos .. lenfile - lenline]
    currentindex[linepos .. linepos + lenline - 1] = index

end procedure

--------------------------------------------------------------------------------

global procedure replaceFileLine(sequence rawtext, integer linepos)
    -- replace line at 'linepos' into current file and update line index

    integer disp, newlen, oldlen, longlen, start, finish
    sequence newline

    -- prepare newline for insertion
    newline = TextToAnalysedLine(rawtext) -- make analysed line
    newline = AnalysedToPackedLine(newline) -- compact into 'lengthed' packed line

    -- prepare line insertion parameters
    start = currentindex[linepos]
    finish = currentindex[linepos+1]
    newlen = length(newline)
    oldlen = finish - start
    longlen = length(files[currentfilepos])
    disp = newlen - oldlen

    -- insert line based on sgn relationship
    if disp != 0 then -- length is not same, so resize gap & update index

        if disp < 0 then -- SHRINKING REQUIRED

            if start < longlen - finish then -- move beginning of sequence
                files[currentfilepos][1 - disp .. start - disp - 1] =
                    files[currentfilepos][1 .. start - 1]
                files[currentfilepos] = files[currentfilepos][1 - disp .. longlen]

            else -- move end of sequence
                files[currentfilepos][finish + disp .. longlen + disp] =
                    files[currentfilepos][finish .. longlen]
                files[currentfilepos] = files[currentfilepos][1 .. longlen + disp]

            end if

        else -- EXPANSION REQUIRED

            if start < longlen - finish then -- move beginning of sequence

                files[currentfilepos] = repeat(0, disp) & files[currentfilepos]
                files[currentfilepos][1 .. start] =
                    files[currentfilepos][1 + disp .. start + disp]

            else -- move end of sequence
                files[currentfilepos] &= repeat(0, disp)
                files[currentfilepos][finish + disp .. longlen + disp] =
                    files[currentfilepos][finish .. longlen]

            end if
    
        end if

         -- common to exp/shr but not necesary for disp=0
        currentindex[linepos+1..length(currentindex)] += disp
    end if

    -- assign new line to vacant space (common to all types of translation)
    files[currentfilepos][start .. start + newlen - 1] = newline

end procedure
--------------------------------------------------------------------------------

procedure refreshFileIndex()
    if currentfilepos then -- greater than 0 so we can use it
        currentindex = makeFileIndex(files[currentfilepos])
    end if
end procedure


--------------------------------------------------------------------------------

global procedure selectOpenFile(integer filepos) -- point indicators to specified opened file

    -- what about no files open (as in case of deleting lone open file)?
    if filepos = 0 then
        resetFileVariables()
        resetActionList()
        updateUndoRedo()
        return
    end if

    -- point to new file
    currentfilepos = filepos
    updateUndoRedo()
    if filepos=1 then
        updateMainMenu(w32True) --enable
--		call_proc(UpdateMainMenu,{w32True})
    end if  

    -- update line index
    refreshFileIndex()

end procedure

--
-- List of entries for the Window menu.
-- If alpha or tab order, each entry is {0, name, hndl, i}.
-- If last mod, each entry is {date, name, hndl, i}, where date
-- is a 6-element sequence from dir() {year,month,day,hour,min,sec},
-- held negative to put most recent first.
--
sequence WinEntries
         WinEntries={}

procedure onClickWinMenu(integer self, integer event, sequence params)
    if event or length(params) then end if  -- suppress warnings
    for i = 1 to length(WinEntries) do
        if self = WinEntries[i][3] then
            call_proc(SelectScreenFile,{WinEntries[i][4]})
            exit
        end if
    end for
end procedure

global procedure rebuildWindowMenu()
    sequence ampersand
    integer hndl
    object date
    if WinMenu and isWindowMenu then        -- editor may need restart...
        --
        -- destroy any WinMenu entries...
        --
        for i = 1 to length(WinEntries) do
            destroy(WinEntries[i][3])
        end for
        if getNumberOfFiles() then
            WinEntries=repeat(0,getNumberOfFiles())
            for i = 1 to length(WinEntries) do
                WinEntries[i]={0,getFileName(i),0,i}
                if isWindowMenu=2 then  --alphabetic insensitive
                    WinEntries[i][2]=lower(WinEntries[i][2])
                end if
                if isWindowMenu=3 then  --most recent
                    date=dir(getFilePath(i)&getFileName(i))
                    if sequence(date) and length(date)=1 then
                        WinEntries[i][1]=-date[1][4..9]
                    end if
                end if
            end for
            if find(isWindowMenu,{1,2,3}) then  -- Alpha or LastMod
                WinEntries=sort(WinEntries)
            end if
            for i = 1 to length(WinEntries) do
                if i<=9 then
                    ampersand={'&','0'+i,' '}       -- "&1 " .. "&9 "
                elsif i<=35 then
                    ampersand={'&','a'+i-10,' '}    -- "&a " .. "&z "
                else
                    ampersand=""
                end if
                hndl=create(MenuItem,ampersand&WinEntries[i][2],WinMenu,0,0,0,0,0)
                setHandler(hndl,w32HClick,routine_id("onClickWinMenu"))
                WinEntries[i][3]=hndl
            end for
        else
            hndl=create(MenuItem,"(empty)",WinMenu,0,0,0,0,0)
            setEnable(hndl,w32False)
            WinEntries={{0,0,hndl,0}}   -- so it gets destroyed
        end if
    end if
end procedure
rebuildWindowMenu()
--------------------------------------------------------------------------------

procedure addOpenFile(sequence s) -- add new encoded file and make it current one
-- only called by loaddiskfile() & newblankfile()
    sequence temp

    if not length(s) then -- completely blank file
        s = {1}
    end if

    -- put data at end of file listings
    files = append(files, s)

    -- create new status structure populated with some default values
    temp = makeNewStatusStructure()

    -- update status
    status = append(status, temp)

    -- update linelengths structure
    linelengths = append(linelengths, {1})

    -- update(reserve space for) actions list
    addActionListForNewFile()

    -- point specifiers to this new file
    selectOpenFile(length(files))

end procedure

--------------------------------------------------------------------------------

global procedure deleteOpenFile(integer filepos)
    -- Delete specified file
    -- If it was the current file then a new current file will be made

    integer len, wascurrent

    -- wascurrent is w32True if current file is the one being deleted
    wascurrent = ( filepos = currentfilepos )

    len = length(files)

    if len <= 1 then -- ascertain if any or only one file is present
        resetFileVariables() -- in which case short-circuit deletion process
        resetActionList()
        updateUndoRedo()
        return
    end if

    -- delete actual data file
    files[filepos .. len - 1] = files[filepos + 1 .. len] -- move upper half down
    files = files[1 .. len - 1] -- trim redundant end

    -- delete state variables
    status[filepos .. len - 1] = status[filepos + 1 .. len]
    status = status[1 .. len - 1]

    -- delete lengths list for that file
    linelengths[filepos .. len - 1] = linelengths[filepos + 1 .. len]
    linelengths = linelengths[1 .. len - 1]

    -- update actionlist
    deleteActionListForFile(filepos)

    -- update file specifiers
    len -= 1
    currentfilepos -= (filepos < currentfilepos)
    if currentfilepos > len then
        currentfilepos = len
    end if
    if wascurrent then
        selectOpenFile(currentfilepos)
    end if
    rebuildWindowMenu()

end procedure

----------------------------------------------------------------------------------------------
global procedure renameOpenFile(integer filepos, sequence newpathname)
    newpathname = extractPathAndName(newpathname)
    status[filepos][FILEPATH] = newpathname[1]
    status[filepos][FILENAME] = newpathname[2]
    rebuildWindowMenu()
end procedure

--------------------------------------------------------------------------------

global procedure makeNewBlankFile() -- create blank file
    addOpenFile( {} )
    status[currentfilepos][SAVEFLAG] = w32False
    status[currentfilepos][FILENAME] = "[untitled]"
    rebuildWindowMenu()
end procedure

--------------------------------------------------------------------------------

global function saveDiskFile(integer filepos, integer forceflag)
-- save opened data file to disk
-- returns w32True for success, w32False for failure
-- checks save flag to determine if saving is really necessary
-- but this can be over-ridden by forceflag = w32True
    integer ID, len
    sequence pathname, s, index

    -- check save flag first
    if forceflag = w32False and status[filepos][SAVEFLAG] = w32False then -- no need to save again
        return w32True
    end if

    if not length(status[filepos][FILEPATH]) then -- no specified path, use default directory
        pathname = current_dir() & '\\'
    else
        pathname = status[filepos][FILEPATH]
    end if
    pathname &= status[filepos][FILENAME]

    ID = open(pathname, "w")
    if ID = -1 then -- can't create file
        return w32False
    end if

    index = makeFileIndex(files[filepos])

    len = length(index) - 1

    --PL 23/09/02
    if isReplaceTabs then
        ExpandTabsToSpaces=1 -- affects AnalysedToTextLine (in analyse.e)
    end if  
    for i = 1 to len do
        s = files[filepos][index[i]+1..index[i+1] -1] -- do not alter this line, logic is stable
        s = PackedToAnalysedLine(s)
        s = TranslateLineUserToEuphoria(s)
        if len = i or len = 1 then -- do not put newline char on end of very last line
            puts(ID, s)
        else
            puts(ID, s & '\n')
        end if
    end for
    ExpandTabsToSpaces=0
    close(ID)

    -- update path & name variables and save flag
    pathname = extractPathAndName(pathname)
    status[filepos][FILEPATH] = pathname[1]
    status[filepos][FILENAME] = pathname[2]
    status[filepos][SAVEFLAG] = w32False
    rebuildWindowMenu()

    return w32True -- success
end function

global function isFileAlreadyOpen(sequence path, sequence name, integer filepos)
    for i = 1 to length(status) do
        if filepos != i then
            if equal(convertToLowerCase(status[i][FILENAME]),convertToLowerCase( name)) then
                if equal(convertToLowerCase(status[i][FILEPATH]), convertToLowerCase(path)) then
                    return 0 - i -- both are same so we are trying to open an already opened file
                end if
            end if
        end if
    end for
    return w32False
end function
--------------------------------------------------------------------------------
function read_in_disk_data(integer openedfileid, sequence name)

    integer len
    sequence result, sLineLengths
    object c

    result = {}
    sLineLengths = {1}

    -- get file extension
    name = extractNameAndExtn(name)
    name = name[2]

    -- at this point the sequence 'name' has the file ext'n
    -- you can test this to alter the behaviour of the way the
    -- file is read in

    while 1 do

    
--------------------------------------------------- MIKE, edit 'tween these lines
        c = gets(openedfileid)

        -- has end of file been reached?
        if integer (c) then
            exit
        end if

         -- remove end_of_line char if needed
        if length(c) and c[length(c)] = '\n' then
            c = c[1..length(c) - 1]
        end if

---------------------------------------------------
        -- 'clean' line of non-ascii chars
        c = filterLine(c)

        -- process length of lines
        len = LinePrintLength(c)
        len += 1 -- to account for any 0 length lines crashing euphorias 1-based system
        if len > length(sLineLengths) then
            sLineLengths &= repeat(0, len - length(sLineLengths))
        end if
        sLineLengths[len] += 1

        -- convert to compressed form
        c = TranslateLineEuphoriaToUser(c)
        c = TextToAnalysedLine(c)
        c = AnalysedToPackedLine(c)

        -- tack on end of sequence
        result &= c

    end while

    return {result, sLineLengths}

end function

global function loadDiskFile(sequence pathname)
-- retrieve disk file into files area
    integer id
    sequence path, name, sLineLengths
    object result

-- return messages:
-- 0 file could not be opened
-- 1 file successfully opened (and made current)
-- -x for file already opened

    -- separate path & name elements from target string ref
    result = extractPathAndName(pathname)
    path = result[1]
    name = result[2]

    -- check if no path given and do something about it if none there
    if not length(path) then -- no path given only filename
        path = current_dir() & '\\' -- so make default directory
    end if

    -- check to see whether it's already open in the editor
    id = isFileAlreadyOpen(path, name, 0)
    if id != w32False then
        return id -- sent as -x to indicate which file to switch to
    end if

    -- prepare fully qualified target
    pathname = path & name

    -- attempt to retrieve data from disk
    result = {}

    -- try to open file
    id = open(pathname, "r")

    -- check if file could not be opened
    if id = -1 then -- error - file could not be opened
        return 0
    end if

    -- get file data + line_lengths
    result = read_in_disk_data(id, name)
    -- returns : {result, sLineLengths}
    
    close(id)

    -- assign to correct variables
    sLineLengths = result[2]
    result = result[1]

    -- file successfully loaded so add to pool of opened files
    addOpenFile(result)
    status[currentfilepos][FILEPATH] = path
    status[currentfilepos][FILENAME] = name
    status[currentfilepos][SAVEFLAG] = w32False
    linelengths[currentfilepos] = sLineLengths
    rebuildWindowMenu()

    -- return success message
    return 1
end function

global function reloadDiskFile(integer filepos)
-- retrieve disk file into files area
    integer id
    sequence path, name, sLineLengths
    object result
    
-- return messages:
-- w32False: file could not be re-opened
-- w32True : file successfully re-opened

    -- prepare variables to capture data
    path = getFilePath(filepos)
    name = getFileName(filepos)
    result = {}

    -- try to open file
    id = open(path & name, "r")

    -- check if file could not be opened
    if id = -1 then -- error - file could not be opened
        return w32False

    else -- reset system values for the file

        -- blank status except name & path
        status[filepos] = makeNewStatusStructure()
        status[filepos][FILEPATH] = path
        status[filepos][FILENAME] = name

        -- blank file data
        files[filepos] = 0

        -- blank line data
        linelengths[filepos] = 0

        -- blank action data
        resetActions(filepos)

    end if

    -- get file data + line_lengths
    result = read_in_disk_data(id, name)
    -- returns : {result, sLineLengths}

    close(id)
    
    -- assign to correct variables
    sLineLengths = result[2]
    result = result[1]

--DEV PL 13/11/02 copied from addOpenFile():
    if not length(result) then -- completely blank file
        result = {1}
    end if
    
    -- now, add file data & line data to system
    files[filepos] = result
    status[filepos][SAVEFLAG] = w32False
    linelengths[filepos] = sLineLengths

    -- return success
    return w32True

end function

global procedure shuffleFile(integer direction)
--
-- swaps the currentfilepos info with that of file +/-1
--
    object temp
    integer c, cd
    c=currentfilepos
    cd=currentfilepos+direction

    temp=status[c] status[c]=status[cd] status[cd]=temp
    temp=files[c] files[c]=files[cd] files[cd]=temp
    temp=linelengths[c] linelengths[c]=linelengths[cd] linelengths[cd]=temp

    currentfilepos+=direction

end procedure   