without trace

-------------------------------------------------------------------------------
---------------------------------PIXMAP BUFFER---------------------------------
-------------------------------------------------------------------------------

-- This code will manage the Pixmap buffer that will be created for the screen.


integer pixmap, -- this will be the main pixmap holding the entire screen
        pix -- this will hold a single line of pixmap info

integer fontwidth, fontheight -- holds details of the font sizes

integer pixmaprows, pixmapcolumns

fontwidth = 1 fontheight=1
pixmaprows= 0 pixmapcolumns = 0

global function getPixmapID()
    return pixmap
end function

-- initialize pixmap

procedure initializePixmap()
    pixmap = create(Pixmap, "", 0, 0, 0, 10, 10, 0)
    pix = create(Pixmap, "", 0, 0, 0, 10, 10, 0)

    if pixmap = 0 or pix = 0 then
        puts(1,"One or both pixmaps uncreated.\n\n")
        ? pixmap
        ? pix
        ?9/0    -- [leave in]
    end if
end procedure

initializePixmap()

-------------------------------------------------------------------------------

global function getFontWidth()
    return fontwidth
end function

global function getFontHeight()
    return fontheight
end function

------------------------------------------------------------------------------

global function resizePixmap(integer width, integer height, sequence fontname, integer fontsize, integer isFontBold)

    sequence junk

    -- set the font into both pixmaps
    if isFontBold = w32True then
        isFontBold = Bold
    else
        isFontBold = Normal
    end if
    setFont(pixmap, fontname, fontsize, isFontBold)
    setFont(pix, fontname, fontsize, isFontBold)

    -- get width & height of font and set into local variables
    junk = getFontSize(pixmap)
    fontwidth = junk[1]
    fontheight = junk[2]

    -- work out the character rows & columns
    pixmapcolumns = floor((width - 1) / fontwidth) --+ 1
    pixmaprows = floor((height - 1) / fontheight) --+ 1

    -- ensure a minimum of 2 rows & 2 coloumns maintained
    if pixmaprows < 2 then
        pixmaprows = 2
    end if
    if pixmapcolumns < 2 then
        pixmapcolumns = 2
    end if

    -- calculate actual pixmap size
    width = pixmapcolumns * fontwidth
    height = pixmaprows * fontheight

    -- resize both pixmap buffers
    setCtlSize(pixmap, width, height)   -- full screen
    setCtlSize(pix, width, fontheight)  -- single line

    -- exit to calling routine with new sizes
    return {pixmapcolumns, pixmaprows, width, height}

end function

global procedure    setFont_specs(sequence fontname, integer fontsize, integer isFontBold)
    setFont(pixmap, fontname, fontsize, isFontBold)
    setFont(pix, fontname, fontsize, isFontBold)
end procedure
-----------------------------------------------------------------------------------------

global procedure printText(integer id, sequence comboblock, integer x, integer y, integer fontwidth, integer fontheight )

    integer tempx
    atom hdc, result, pstr
    sequence BBox, comboline, textline, text, colourline, backcolour

    -- early out?
    if length( comboblock ) = 0 then
        return
    end if

    -- get the device context
    hdc = assignFont( id)

    -------------------
    -------------------
    -- display the text
    -------------------
    -------------------


    tempx = x

    for i = 1 to length(comboblock) do

        -- extract data for the current line
        comboline = comboblock[i]

        textline = comboline[1]
        colourline = comboline[2]
        backcolour = comboline[3]

        -- save x posn
        x = tempx

        for j = 1 to length(textline) do

            text = textline[j]

            -- set the Pen Position
            BBox = convPctToPixel(id, x,y, 0, 0)

            -- set the text & background color for the window
            result = w32Func( xSetTextColor, {hdc, colourline[j] } )
            result = w32Func( xSetBkColor, {hdc, backcolour[j] } )
            result = w32Func( xSetBkMode, {hdc, OPAQUE } )

            -- copy string to memory
            pstr = w32acquire_mem(0, text )

            -- write out string to window/device
            w32Proc( xTextOut, { hdc, BBox[1], BBox[2], pstr, length( text ) } )

            -- advance column pointer to next start pos
            x += length(text) * fontwidth

            -- Free the string
            release_mem(pstr)

        end for

        -- advance row to next position
        y += fontheight

    end for

    -- release the device context
    releaseDC( id )

end procedure

global procedure printToPixmap(integer X, integer Y, integer XX, integer YY)
    -- print the block of coloured text from the buffer at the coordinates specified

    sequence block

    -- blank zone that data will be printed to
    setPenColor(pixmap, getColour(COLOURBACKGROUND) )
        drawRectangle(pixmap, 1,
            X * fontwidth - fontwidth, Y * fontheight - fontheight,
            XX * fontwidth,
            YY * fontheight)

    -- get the coloured block that will be printed
    block = makeBufferPrintBlock(Y, X, YY, XX)

    -- set initial X & Y coords
    X = X * fontwidth - fontwidth
    Y = Y * fontheight - fontheight

    -- call new routine in win32lib.ew to print multi-lines of coloured text
    printText( pixmap, block, X, Y, fontwidth, fontheight )

end procedure

-------------------------------------------------------------------------------

procedure movePixmapLine (integer destCol, integer destRow, integer srcCol, integer srcRow, integer srcLen)
-- using pix (aka pix) move a line around on the main pixmap

    integer width

    width = fontwidth * srcLen

    -- copy to pix buffer
    bitBlt
        (pix,               -- destination pixmap
        0,                  -- x coord of destination
        0,                  -- y coord of destination
        pixmap,                             -- source
        srcCol * fontwidth - fontwidth,     -- x coord of source
        srcRow * fontheight - fontheight,   -- y coord of source
        width,                              -- width of source
        fontheight,         -- height of source
        SrcCopy )           -- raster operation

    -- copy back to pixmap in new position
    bitBlt
        (pixmap,                            -- destination pixmap
        destCol * fontwidth - fontwidth,    -- x coord of destination
        destRow * fontheight - fontheight,  -- y coord of destination
        pix,                                -- source
        0,                                  -- x coord of source
        0,                                  -- y coord of source
        width,                              -- width of source
        fontheight,                         -- height of source
        SrcCopy )                           -- raster operation

end procedure


global procedure movePixmapBlock(integer srcCol, integer srcRow, integer width, integer height, integer destCol, integer destRow)
 -- move a block of data around on the main pixmap

    -- transfer lines of pixmap data one at a time accounting for the direction of the transfer
    if destRow <= srcRow then
        for i = 0 to height - 1 do
            movePixmapLine( destCol, destRow + i, srcCol, srcRow + i, width)
        end for
    else
        for i = height - 1 to  0 by -1 do
            movePixmapLine( destCol, destRow + i, srcCol, srcRow + i, width)
        end for
    end if

end procedure


----------------------------------------------------------------------

global procedure deletePixmapLines(integer cavitytop, integer cavitybottom)
-- remove the block referred to and scroll up the block from lower half of screen
-- NB: lowerhalf is NOT overwritten by this routine

    -- transfer the block below cavitybottom up to cavitytop
    movePixmapBlock
        (1, cavitybottom + 1, pixmapcolumns, pixmaprows - cavitybottom, 1, cavitytop)

end procedure

-------------------------------------------------------------------------------

global procedure insertPixmapLines(integer cavitytop, integer cavitybottom)
-- move lower half of these buffers down opening up a gap for other lines to be inserted
-- NB: mid-block is not overwritten with new text by this routine

    -- transfer the mid block into the lower half of the pixmap
    movePixmapBlock
        (1, cavitytop, pixmapcolumns, pixmaprows - cavitybottom, 1, cavitybottom + 1)

end procedure

-------------------------------------------------------------------------------

global procedure printPixmapLines(integer top, integer bottom)
-- using the update info in the buffers print to the current pixmap

    printToPixmap( 1, top, pixmapcolumns, bottom)

end procedure

-------------------------------------------------------------------------------

global procedure transferPixmapLinesToWindow( integer top, integer bottom)
-- transfer specified lines from the pixmap to the screen
    bitBlt(
        MAIN,           -- destination pixmap
        editx,          -- x coord of destination
        edity + top * fontheight - fontheight,-- y coord of destination
        pixmap,         -- source
        0,              -- x coord of source
        top * fontheight - fontheight,  -- y coord of source
        pixmapcolumns * fontwidth,      -- width of source
        (bottom - top + 1) * fontheight,-- height of source
        SrcCopy )       -- raster operation
                    
end procedure

-------------------------------------------------------------------------------

global procedure transferPixmapToWindow() -- transfer entire pixmap to the screen

    -- if no files opened then print blank screen
    if getNumberOfFiles() = 0 then
        setPenColor(MAIN, getSysColor(COLOR_BTNFACE))
        drawRectangle(MAIN, 1,
            editx, edity,
            editx + pixmapcolumns * fontwidth - 1,
            edity + pixmaprows* fontheight - 1)
        return

    else -- copy whole of pixmap to the screen
        transferPixmapLinesToWindow( 1, pixmaprows)

    end if

end procedure

-------------------------------------------------------------------------------