
-- code to quickly extract names of all routines in the current file
-- provision made for getting scope & type

global function max(integer a, integer b)
    if a>b then return a else return b end if
end function
global function min(integer a, integer b)
    if a<b then return a else return b end if
end function

-------------------------------------------------------------
--				routine list window

constant
    ROUTINEWINDOW = create( Window, "Quick Jump", MAIN, 100, 10, 500, 410, 0),
    All = create( Radio, "&All", ROUTINEWINDOW, 8, 15, 42, 20, 0 ),
    Globals = create( Radio, "&Globals", ROUTINEWINDOW, 58, 15, 66, 20, 0 ),
    Sections = create( Radio, "&Sections", ROUTINEWINDOW, 138, 15, 72, 20, 0 ),
    Filtxt = create( LText, "&Filter", ROUTINEWINDOW, 220, 15, 35, 20, 0 ),
    Filter = create( EditText, "", ROUTINEWINDOW, 255, 12, 237, 32, 0 ),
    ROUTINELIST = create( List, "", ROUTINEWINDOW, 10, 40, {w32Edge,-10}, {w32Edge,-10}, 0 )

    setCheck(All, w32True)
    setCheck(Globals, w32False)
    setCheck(Sections, w32False)

    setVisible(ROUTINEWINDOW, w32False)
    setFont(ROUTINELIST, "Courier New", getSizeOffont(), 0 )

    setWindowBackColor(ROUTINELIST,getColour(COLOURBACKGROUND))

function populateRoutineList(sequence names)
    sequence Ftext
    integer tally
    eraseItems ( ROUTINELIST )  -- remove prev details if any
    -- populate list, applying filter
    tally=0
    Ftext=lower(getText(Filter))
    for i = 1 to length(names) do
        if not length(Ftext)
        or match(Ftext,lower(names[i])) then
            addItem( ROUTINELIST, names[i] )
            tally+=1
        end if
    end for
    return tally
end function

procedure onResizeWindow(integer self, integer event, sequence parms)
    sequence size
    if self or event or length(parms) then end if   -- suppress warnings
    size = getClientRect(ROUTINEWINDOW) -- returns {x1,y1,x2,y2}
    size[3]-=size[1]
    size[4]-=size[2]
    setCtlSize(ROUTINELIST,size[3]-18,size[4]-45)
    setCtlSize(Filter,size[3]-263,20)
end procedure
setHandler(ROUTINEWINDOW, w32HResize, routine_id( "onResizeWindow" ))

global procedure closeROUTINEWINDOW(integer self, integer event, sequence params)
    if self or event or length(params) then end if  -- suppress warnings
    setVisible(ROUTINEWINDOW, w32False)
    setFocus(MAIN)
end procedure
setHandler(ROUTINEWINDOW,w32HClose,routine_id("closeROUTINEWINDOW"))

sequence routinescope, routinetype, routinenames, routinelinenumbers
sequence globalroutinenames, globalroutinelinenumbers
sequence sectionnames, sectionlinenumbers

function getActualLineNumberOfRoutine(integer ind)
    sequence names, linenumbers, Ftext
    if isChecked(All) then
        names = routinenames
        linenumbers = routinelinenumbers
    elsif isChecked(Globals) then
        names = globalroutinenames
        linenumbers = globalroutinelinenumbers
    else
        names = sectionnames
        linenumbers = sectionlinenumbers
    end if

    Ftext=lower(getText(Filter))
    if not length(Ftext) then
        return linenumbers[ind]
    end if
    for i = 1 to length(names) do
        if match(Ftext,lower(names[i])) then
            ind-=1
            if ind=0 then return linenumbers[i] end if
        end if
    end for
    if message_box("Internal Error","Sorry",MB_OK) then end if
    return 1    -- just jump to line one then
end function

procedure jumpToNewLine(integer newline)
--DEV ideally this should not force to line 1, as per find kludge?
    moveToNewEditLine(newline) -- PL 27/09/02
    setCursorPosition(newline, 1)
    if jumpToNewScreenLine(newline) then
        transferPixmapToWindow()    -- as displayScreen()
        setScrollPositions()
    end if
    showCursor()
    setFocus( MAIN )
end procedure

procedure onMouse_ROUTINELIST(integer self, integer event, sequence params)
integer index
    if self or event then end if -- suppress warnings
    if params[1] = LeftDoubleClick then
        index = getIndex(ROUTINELIST)
        if index > 0 then
            closeROUTINEWINDOW(0,0,{})
            index = getActualLineNumberOfRoutine(index)
            jumpToNewLine(index)
        end if
    end if
end procedure
setHandler(ROUTINELIST,w32HMouse,routine_id("onMouse_ROUTINELIST"))


constant
    endword = find("end", keywords),
    globalword = find("global", keywords),
    typeword = find("type", keywords),
    functionword = find("function", keywords),
    procedureword = find("procedure", keywords)

procedure resetRoutineLists()
    routinescope = {}
    routinetype = {}
    routinenames = {}
    routinelinenumbers = {}

    globalroutinelinenumbers = {}
    globalroutinenames = {}

    sectionlinenumbers = {}
    sectionnames = {}
end procedure

function getRoutineTypeIndex(sequence cand)
    if cand[1] then
        return 1 -- TYPE
    elsif cand[2] then
        return 2 -- FUNCTION
    elsif cand[3] then
        return 3 -- PROCEDURE
    else
        ? 9/0   -- [leave in] major error
    end if
end function

function getRoutineName( sequence aline)
    integer len, start, finish

    aline = AnalysedToTextLine(aline)
    len = length(aline)
    start = 0

    for i = 1 to len do
        if isLetter(aline[i]) then
            start = i
            exit
        end if
    end for

    if start = 0 then
        return ""
    end if

    finish = len

    for i = start to len do
        if not (isLetter(aline[i]) or isNumber(aline[i]) or aline[i] = '_') then -- end found
            finish = i - 1
            exit
        end if
    end for

    return aline[start .. finish]

end function


procedure extractAnyRoutineInformation(sequence aline, integer lnumber)
    integer glo, typ, fun, pro
    sequence name, temp


    temp = unpackChars(aline[1])

    glo = find(globalword, temp)

    typ = 0
    fun = 0
    pro = 0
    name = ""

    while 1 do

        -- ensure that closing of block does not fool program
        if find(endword, temp) then
            exit
        end if

        typ = find(typeword, temp)
        if typ then
            aline = PackedToAnalysedLine(aline)
            name = getRoutineName(aline[typ + 1 .. length(aline)])
            exit
        end if

        fun = find(functionword, temp)
        if fun then
            aline = PackedToAnalysedLine(aline)
            name = getRoutineName(aline[fun + 1 .. length(aline)])
            exit
        end if

        pro = find(procedureword, temp)
        if pro then
            aline = PackedToAnalysedLine(aline)
            name = getRoutineName(aline[pro + 1 .. length(aline)])
            exit
        end if

        name = ""
        exit

    end while

    if equal(name, "") then
        return -- failed to get name
    end if

    routinescope &= glo
    routinetype &= getRoutineTypeIndex({typ, fun, pro})
    routinenames = append(routinenames, name)
    routinelinenumbers &= lnumber

end procedure

integer isExErrFile isExErrFile=0

procedure cycleThroughAllLines()
    sequence line
    integer c
    sequence prevline,temp

    prevline=""
    for i = 1 to getNumberOfLinesInFile() do
        line = getPackedLine(i)
        c = line[1] -- this is length integer
        if c > 1 then -- possible candidate
            extractAnyRoutineInformation(line[2..length(line)], i)
        end if
        if c > 2    -- possible section underline
        and (length(prevline) or isExErrFile) then
            if isExErrFile then
            temp = unpackChars(line[2])
            if find(temp[3],".\\:") then
                line = PackedToAnalysedLine(line[2..length(line)])
                line = AnalysedToTextLine(line)
                sectionnames=append(sectionnames,line)
                sectionlinenumbers&=i
            end if
            
            else
            if equal(line[c-1],609073776)       -- "===="               
            or equal(line[c-1],449316720) then  -- "----"
                prevline = PackedToAnalysedLine(prevline[2..length(prevline)])
                while length(prevline) and (prevline[1]=31 or prevline[1]=32) do --tab,space
                    prevline=prevline[2..length(prevline)]
                end while
                if compare(prevline,{2,32,4})!=0        -- not exactly "end procedure"
                and compare(prevline,{2,32,17})!=0 then -- not exactly "end function"
                    prevline = AnalysedToTextLine(prevline)--[2..length(prevline)])
                    for k = 1 to length(prevline) do
                        if prevline[k]>='1' and prevline[k]<='z' then
                            sectionnames=append(sectionnames,prevline[k..length(prevline)])
                            sectionlinenumbers&=i-1
                            exit
                        end if
                    end for
                end if
            end if
            end if
        end if
        prevline=line
    end for
end procedure

global procedure setListOfAllRoutines()
    -- clear public list vars
    resetRoutineLists()

    -- extract routine info
    cycleThroughAllLines()

    -- prepare list of global routines
    for i = 1 to length(routinescope) do
        if routinescope[i] then
            globalroutinenames = append(globalroutinenames, routinenames[i])
            globalroutinelinenumbers &= routinelinenumbers[i]
        end if
    end for
end procedure

----------------

function getListRelPlace(integer cur_line)
    integer ind, len
    sequence linenumbers, names, Ftext

    -- get index of ALL routines or just GLOBAL ones
    if isChecked(All) then
        names = routinenames
        linenumbers = routinelinenumbers
    elsif isChecked(Globals) then
        names = globalroutinenames
        linenumbers = globalroutinelinenumbers
    else
        names = sectionnames
        linenumbers = sectionlinenumbers
    end if
    len = length(linenumbers)

    -- get position of where cur_line would be in list
    ind = 0
    Ftext=lower(getText(Filter))
    for i = length(linenumbers) to 1 by -1 do
        if (not length(Ftext) or match(Ftext,lower(names[i])))
        and linenumbers[i]<=cur_line then
            ind+=1
        end if      
    end for

    if ind = 0 then
        return {0, 0}
    elsif ind <= 10 then
        return {ind, 0}
    end if
    return {ind,max(1,ind-5)}
end function

procedure ongotfocusFiltxt(integer self, integer event, sequence params)
    if self or event or length(params) then end if  -- suppress warnings
    setFocus(Filter)
end procedure
setHandler(Filtxt,w32HGotFocus,routine_id("ongotfocusFiltxt"))

procedure onkeydowninlist(integer self, integer event, sequence params)
    integer keycode, index
    if event then end if -- suppress warnings
    keycode=params[1]
    if keycode = VK_ENTER then
        if (self = ROUTINELIST) then
            index = getIndex(ROUTINELIST)
            closeROUTINEWINDOW(0,0,{})
            if index > 0 then
                index = getActualLineNumberOfRoutine(index)
                jumpToNewLine(index)
            end if
            return
        end if
    end if
    if keycode = VK_ESCAPE then
        closeROUTINEWINDOW(0,0,{})
        return
    end if
end procedure
setHandler({All,Globals,Sections,Filter,ROUTINELIST},w32HKeyDown,routine_id("onkeydowninlist"))


procedure displayNames(integer self, integer event, sequence params)
    sequence names, ind_pos
    if event or length(params) then end if  -- suppress warnings
    -- get the names
    if isChecked(All) then
        names = routinenames
    elsif isChecked(Globals) then
        names = globalroutinenames
    else
        names = sectionnames
    end if

    -- insert these into list
    if not populateRoutineList(names) and self!=Filter then
        setText(Filter,"")
        if populateRoutineList(names) then end if
    end if

    -- display the window
    if isVisible(ROUTINEWINDOW)=w32False then
        openWindow(ROUTINEWINDOW, Normal)   
    end if
    if self!=Filter then
        setFocus(ROUTINELIST)
        -- set scroll index to be in relative to cursor place in file
        ind_pos = getListRelPlace(getCursorLine())
        setIndex(ROUTINELIST, ind_pos[2])   -- move to reasonable position
        setIndex(ROUTINELIST, ind_pos[1])   -- then select right line mid-screen
    end if
end procedure
setHandler({All,Globals,Sections},w32HClick,routine_id("displayNames"))
setHandler(Filter,w32HChange,routine_id("displayNames"))

global procedure onClickToolQjmp(integer self, integer event, sequence params)
    sequence junk
    if self or event or length(params) then end if  -- suppress warnings
    -- early out
    if not getNumberOfFiles() then
        return
    end if

    -- switch to Sections view for non-Euphoria files
    if isVisible(ROUTINEWINDOW)=w32False then
        junk=getVal(FILENAME)
        isExErrFile=equal(lower(junk),"ex.err")
        for i = length(junk) to 1 by -1 do
            if junk[i]='.' then
                junk=upper(junk[i+1..length(junk)])
                if find(junk,{"EXW","EW","E","EX","EXU","EU"}) then
                    setCheck(All,w32True)
                    setCheck(Globals,w32False)              
                    setCheck(Sections,w32False)
                else
                    setCheck(All,w32False)
                    setCheck(Globals,w32False)
                    setCheck(Sections,w32True)
                end if
                exit
            end if
        end for
    end if
    
    -- prepare list of routines names, scope & type (& linenumbers)
    setListOfAllRoutines()

    displayNames(0,0,"")

end procedure
setHandler(ToolQJmp,w32HClick,routine_id("onClickToolQjmp"))

global function isLocalRoutine(sequence word)
--
-- Returns zero if the word is unrecognised, else the line number of the
-- routine definition in the current file.
--
    integer idx
    if getNumberOfFiles() then
        -- prepare list of routines names, scope & type (& linenumbers)
        setListOfAllRoutines()
        idx=find(word,routinenames)
        if idx then
            return routinelinenumbers[idx]
        end if
    end if
    return 0
end function