without trace
-------------------------------------------------------------------------------
-----------------------------manage screen-------------------------------------
-------------------------------------------------------------------------------

integer topline, leftcolumn, screencolumns, screenrows

sequence fontname
integer fontsize
integer isFontBold

topline = 1
leftcolumn = 1
screencolumns = 1
screenrows = 1

global atom findflag    -- set to 0.3 when called from/because of find op.
findflag=1              -- for better repositioning, else set to 1
                        -- PL 10/11/02 now also used for F12 jump to error and 
                        -- from the Examine Code button on context sensitive help (F1)

global function getFontName()
    return fontname
end function

global function getSizeOffont()
    return fontsize
end function

global function getFontBold()
    return isFontBold
end function

global procedure setFontSize(integer size)
    fontsize = size
end procedure

global procedure setFontBold(integer bold_flag)
    isFontBold = bold_flag
end procedure


-------------------------------------------------------------------------------
global procedure storeFontSettingsToIni()
    switchToIniGroup("edit")
    setIniValue("font name", fontname)
    setIniValue("font size", fontsize)
    setIniValue("isFontBold", isFontBold)
end procedure
    
procedure initializeFontSettings()

    -- select correct ini group
    switchToIniGroup("edit")

    -- define entry names/headings
    defineIniHeading("font name", TEXT)
    defineIniHeading("font size", INT)
    defineIniHeading("isFontBold", BOOL)

    -- get values and place in variables
    -- FONT NAME
    fontname = getIniValue("font name", "Courier New")
    fontname = "Courier New" -- insist that this font is used first off for now

    -- FONT SIZE
    fontsize = getIniValue("font size", 10)
    if fontsize < 5 or fontsize > 20 then
        fontsize = 10
    end if

    -- FONT BOLDNESS
    isFontBold = getIniValue("isFontBold", w32False)

    setFont_specs(fontname, fontsize, isFontBold)

end procedure

initializeFontSettings()

-----------------

global function getTopLine()
    return topline
end function

global procedure setTopLine(integer top)
    topline = top
end procedure

-----------------

global function getLeftColumn()
    return leftcolumn
end function

global procedure setLeftColumn(integer left)
    leftcolumn = left
end procedure

-----------------

global function getScreenColumns()
    return screencolumns
end function

global function getScreenRows()
    return screenrows
end function

------------------------------------------------------------------------------

function clip_top(integer top)

    -- convert to make relative to screen
    top -= topline - 1

    -- clip lines outside screen
    if top < 1 then
        top = 1
    end if

    return top

end function

----------------

function clip_bottom(integer bottom)

    -- convert to make relative to screen
    bottom -= topline - 1

    -- clip lines outside screen
    if bottom > screenrows then
        bottom = screenrows
    end if

    return bottom

end function

-------------------------------------------------------------------------------

global procedure displayScreenLines(integer top, integer bottom)

    -- make relative to screen
    top = clip_top(top)
    bottom = clip_bottom(bottom)

    -- early out?
    if top > bottom then
        return
    end if

    -- transfer to screen
    transferPixmapLinesToWindow( top, bottom)

end procedure

-------------------------------------------------------------------------------

global procedure setScreenLines(integer top, integer bottom)

    -- make relative to screen
    top = clip_top(top)
    bottom = clip_bottom(bottom)

    -- early out?
    if bottom < top then -- selection falls entirely outside the screen
        return
    end if

    -- update buffers
    setBufferLines(top, bottom)
    highlightBufferLines(top, bottom)
    printPixmapLines(top, bottom)

end procedure

-------------------------------------------------------------------------------

global procedure highlightScreenLines(integer top, integer bottom)
    -- make relative to screen
    top = clip_top(top)
    bottom = clip_bottom(bottom)

    -- early out?
    if bottom < top then -- selection falls entirely outside the screen
        return
    end if

    highlightBufferLines(top, bottom)
    printPixmapLines(top, bottom)
    setDisplayRange(top, bottom)

end procedure

-------------------------------------------------------------------------------

global procedure unhighlightScreen()
-- remove any highlighting in the buffers and on the screen
    integer top, bottom

    top = getSelectTop()
    bottom = getSelectBottom()

    resetSelection()

    if top = 0 then -- no highlighting to undo
        return
    end if

    setDisplayRange(top, bottom)

    top = clip_top(top)
    bottom = clip_bottom(bottom)

    -- early out?
    if bottom < top then -- selection falls entirely outside the screen
        return
    end if

    if bottom > getBufferLength() then
        bottom = getBufferLength()
    end if

    unhighlightBufferLines(top, bottom)
    printPixmapLines(top, bottom)


end procedure

-------------------------------------------------------------------------------


global procedure insertScreenLines(integer fTop, integer fBottom)
-- (Scroll)open a block of lines in the screen and insert fresh lines from file
-- NB: bottom line is not scrolled open as it will be the line on which the cursor is
-- and should remain unchanged
    integer top, bottom

    -- assign params to scratch vars to avoid damaging TOPLINE if that will alter
    top = fTop
    bottom = fBottom

    -- clip top & bottom bounds
    top = clip_top(top)
    bottom = clip_bottom(bottom)

    -- does selection fall entirely outside the screen?
    if top > bottom then
        return
    end if

    -- make space for the new lines entering the buffers
    insertBufferLines(top, bottom - 1)
    insertPixmapLines( top, bottom - 1 )

    -- re-appoint the displaced lines
    setScreenLines(top + topline - 1, bottom + topline - 1)

    if topline > fTop then -- must update this value to avoid crash
        topline += fBottom - fTop + 1
    end if

end procedure



-------------------------------------------------------------------------------

global procedure deleteScreenLines (integer fTop, integer fBottom) -- delete the specified lines

    integer top, bottom

    -- exit if lines selected fall entirely outside screen BUT UPDATE new TOPLINE
    if fBottom < topline then
        topline -= fBottom - fTop + 1 -- important line!!!
        return
    elsif fTop >= topline + screenrows then
        return
    end if

    -- clip bounds
    top = fTop - topline + 1
    bottom = fBottom - topline + 1

    if top < 1 then
        top = 1
    end if
    topline = fTop - top + 1 -- important line!!!
    if bottom > screenrows then
        bottom = screenrows
    end if

    setBufferTopLine(topline)
    deleteBufferLines(top, bottom)

    -- delete pixmap lines
    deletePixmapLines(top, bottom)

    -- calc beginning (top) of lower half
    top += screenrows - bottom

    -- re-appoint the displaced lines
    setBufferLines( top, screenrows )
    highlightBufferLines( top, screenrows )
    printPixmapLines( top, screenrows )

end procedure
    

-------------------------------------------------------------------------------

global procedure refreshScreen() -- refresh the edit area
    setFont_specs(fontname, fontsize, isFontBold)
    -- update the text & colour buffers
    initializeBuffers(topline, leftcolumn, screenrows, getTabWidth() )

    -- any highlighting needed?
    highlightBuffer()

    -- update the current pixmap with full page of data
    printPixmapLines(1, screenrows)

end procedure

-------------------------------------------------------------------------------


global function resizeScreen (integer width, integer height)
-- resize edit screen but don't print to MAIN window

    sequence junk

    -- resize the pixmaps and get back the new limits
    junk = resizePixmap(width, height, fontname, fontsize, isFontBold)
    --returns {screencolumns, screenrows, width, height}
    -- update local variables with the new information
    screencolumns = junk[1]
    screenrows = junk[2]

    if getNumberOfFiles() then -- ok, have real file to display

        -- resize the text & colour buffers
        resizeBuffers(screenrows, getTabWidth() )

        -- insert relevant highlighting
        highlightBuffer()

        -- print the entire screen of text onto pixmap
        printPixmapLines(1, screenrows)
    end if

    -- send back details of the size of the shown edit area
    return junk[3..4] -- {width, height}

end function

global function reconstructScreen   (integer width, integer height)
-- resize edit screen using refreshed data (no MAIN printing)

    sequence junk

    -- resize the pixmaps and get back the new limits
    junk = resizePixmap(width, height, fontname, fontsize, isFontBold)
    --returns {screencolumns, screenrows, width, height}
    -- update local variables with the new information
    screencolumns = junk[1]
    screenrows = junk[2]

    if getNumberOfFiles() then -- ok, have real file to display

        -- completely refresh the text & colour buffers
        initializeBuffers (topline, leftcolumn, screenrows, getTabWidth() )

        -- insert relevant highlighting
        highlightBuffer()

        -- print the entire screen of text onto pixmap
        printPixmapLines(1, screenrows)

    end if

    -- send back details of the size of the shown edit area
    return junk[3..4] -- {width, height}

end function
-------------------------------------------------------------------------------

global function scrollScreen(integer newtop, integer newleft)

    -- *efficiently* scroll pixmap based on new parameters as above.

    integer d               -- direction flag
    integer v, h            -- vertical & horizontal displacements to make my life easier!!
    integer rows, cols      -- as above but for screenrows & screencolumns

    -- assign scratch vars
    rows = screenrows
    cols = screencolumns

    -- calculate local displacements
    v = w32abs(newtop - topline)
    h = w32abs(newleft - leftcolumn)

    -- completely refresh buffers & pixmap?
    if v > rows - 3 or h > cols - 3 then
        topline = newtop
        leftcolumn = newleft
        refreshScreen()
        return 1
    end if

    -- ensure that printToPixmap() below will correctly work
    setBufferLeftColumn(newleft) -- this done so printToPixmap() will function correctly

    -- scroll the text & colour buffers
    scrollBuffers(newtop)

    -- any highlighting needed?
    highlightBuffer()

    -- calculate index of scroll direction
    if newtop < topline then 
        d = 6
    elsif newtop > topline then 
        d = 0
    else 
        d = 3 
    end if
    if newleft > leftcolumn then 
        d += 2
    elsif newleft = leftcolumn then 
        d += 1
    end if

    -- interpret scroll direction from index and then transfer main block around
    if d = 0 then -- down & left			**
        movePixmapBlock( 1, v + 1, cols - h,    rows - v, h + 1, 1)
    elsif d = 1 then -- down
        movePixmapBlock( 1, v + 1, cols, rows - v, 1, 1)
    elsif d = 2 then -- down & right		**
        movePixmapBlock( h + 1, v + 1, cols, rows - v, 1, 1)
    elsif d = 3 then -- left
        movePixmapBlock( 1, 1, cols - h, rows, h + 1, 1)
    elsif d = 4 then -- no change, so don't do anything
        return 0
    elsif d = 5 then -- right
        movePixmapBlock( h + 1, 1, cols - h, rows, 1, 1)
    elsif d = 6 then -- up & left			**
        movePixmapBlock( 1, 1, cols - h, rows - v, h + 1, v + 1)
    elsif d = 7 then --  up
        movePixmapBlock( 1, 1, cols, rows - v, 1, v + 1)
    elsif d = 8 then -- up & right		**
        movePixmapBlock( h + 1, 1, cols - h, rows - v, 1, v + 1)
    end if

    -- determine which horizontal blocks go where
    if d < 3 then -- translation was down: L/C/R
        printToPixmap( 1, rows - v + 1, cols, rows)
    elsif d = 3 then -- translation was left
        printToPixmap( 1, 1, h, rows)
    elsif d = 5 then -- translation was right
        printToPixmap( cols - h + 1, 1, cols, rows)
    elsif d > 5 then -- translation was up: L/C/R
        printToPixmap( 1, 1, cols, v)
    end if

    -- calculate minor vertical block (if any)
    if d = 0 then
        printToPixmap( 1, 1, h, rows - v)
    elsif d = 2 then
        printToPixmap( cols - h + 1, 1, cols, rows - v)
    elsif d = 6 then
        printToPixmap( 1, v + 1, h, rows)
    elsif d = 8 then
        printToPixmap( cols - h + 1, v + 1, cols, rows)
    end if

    -- update local variables
    topline = newtop
    leftcolumn = newleft

    return 1

end function

-------------------------------------------------------------------------------
global function repositionScreen(integer curLine, integer curCol)
    integer newtop, newleft

    if curLine >= topline and curLine < topline + screenrows and
        curCol >= leftcolumn and curCol < leftcolumn + screencolumns then -- no need to scroll screen
        return w32False
    end if

    -- initialize default values
    newtop = topline
    newleft = leftcolumn

    -- determine vertical movement
    if curLine < topline then
        newtop = curLine

    elsif curLine >= topline + screenrows then
        newtop = curLine - floor(screenrows*findflag) + 1   --PL 18/09/02


    end if
    
    -- determine horizontal movement
    if curCol < leftcolumn then
        newleft = curCol
        if newleft < 1 then
            newleft = 1
        end if

    elsif curCol >= leftcolumn + screencolumns then
        newleft = curCol - screencolumns + 1

    end if

    return scrollScreen(newtop, newleft)

end function

global function repositionScreenForWordSelection()
-- if a 'word' has been selected by either doubleclick or findtext then manipulate screen from here
    integer newtop, newleft
    integer line, leftcol, rightcol

    line = getSelectTop()
    leftcol = getSelectLeft()
    rightcol = getSelectRight()

    -- initialize default values
    newtop = topline
    newleft = leftcolumn

    -- determine vertical movement
    if line < topline then
        newtop = line
    elsif line >= topline + screenrows then
        newtop = line - floor(screenrows*findflag) + 1  -- PL 18/09/02
    end if
    
    -- determine horizontal movement
    if leftcol < leftcolumn then
        newleft = leftcol
        if newleft < 1 then
            newleft = 1
        end if
    end if
    if rightcol + 1 >= leftcolumn + screencolumns then
        newleft = rightcol + 1 - screencolumns + 1
    end if

    return scrollScreen(newtop, newleft)
end function

global function jumpToNewScreenLine(integer curLine)
    -- similar to repositionScreen() but for the purpose of jumping to a new routine
    integer newtop

    newtop = curLine

    if getNumberOfLinesInFile() - curLine < getScreenRows() then
        newtop = getNumberOfLinesInFile() - getScreenRows() + 1
    end if

    if newtop < 1 then
        newtop = 1
    end if

    return scrollScreen(newtop, 1)

end function