-----------------------------------------------------------------------------
---------------------------------tab code------------------------------------
-----------------------------------------------------------------------------
-- code to assist with the tab problem


integer tabwidth -- this will hold the actual value of the tab spacings

defineIniGroup("edit")
defineIniHeading("tab spacing", INT)

tabwidth = getIniValue("tab spacing", 4)

global procedure setTabValuetoIni()
    switchToIniGroup("edit")
    setIniValue("tab spacing", tabwidth)
end procedure

constant TAB = '\t'

global type isLabelCharacter(integer char) -- returns true if char is label character
    return isLetter(char) or char = '_' or isNumber(char)
        --(char >= 'a' and char <= 'z') or
        --(char >= 'A' and char <= 'Z') or
    --	char = '_'
end type

global function getTabWidth()
    return tabwidth
end function

global procedure setTabWidth(integer newtab)
    tabwidth = newtab
end procedure

-----------------------------------------------------------------------------
    
global function PrintToBytePos(integer printpos, sequence rawtext)
-- convert the printpos (print position on screen) to the byte position in the rawtext
-- used to get around the 'tab' problem. I think this is better than simply adding a
-- bunch of spaces to each line
-- it is the complement of the ByteToPrintPos routine
    integer len, junk

    -- early out for 'left=1' or 'right=0'
    if printpos < 2 then
        return printpos
    end if

    len = 0
    junk = 0

    for i = 1 to length(rawtext) do
        junk = rawtext[i]
        if junk = TAB then
            len += tabwidth - remainder(len, tabwidth)
        else
            len += 1
        end if
        if printpos <= len then
            return i
        end if
        junk = i
    end for
    return junk + 1
end function

--------------------

global function ByteToPrintPos(integer bytepos, sequence rawtext)
-- convert the byte position in the rawtext to the printpos (print position on screen)
-- it is the complement of the PrintToBytePos() routine
    integer len, junk
    len = 1
    for i = 1 to length(rawtext) do
        if bytepos = i then
            return len
        end if
        junk = rawtext[i]
        if junk = TAB then
            len += tabwidth - remainder(len-1, tabwidth)
        else
            len += 1
        end if

    end for
    return len
end function

-----------------------------------------------------------------------------

global function LinePrintLength(sequence rawtext)
-- get the total print length of the line
    integer len, junk
    len = 0
    for i = 1 to length(rawtext) do
        junk = rawtext[i]
        if junk = TAB then
            len += tabwidth - remainder(len, tabwidth)
        else
            len += 1
        end if
    end for
    return len
end function

global function getPrintLine(sequence rawtext) -- used to print a line on paper
    sequence r
    integer len, junk, temp
    len = 0
    r = ""
    for i = 1 to length(rawtext) do
        junk = rawtext[i]
        if junk = TAB then
            temp = tabwidth - remainder(len, tabwidth)
            len += temp
            r &= repeat(' ', temp)
        else
            len += 1
            r &= junk
        end if
    end for
    return r
end function
-----------------------------------------------------------------------------

global function PrintCharEndPos(integer printpos, sequence rawtext)
-- get the end pos of the print char whose space includes pos

    integer len, junk

    len = 0

    for i = 1 to length(rawtext) do
        junk = rawtext[i]
        if junk = TAB then
            len += tabwidth - remainder(len, tabwidth)
        else
            len += 1
        end if
        if printpos <= len then exit end if
    end for

    return len

end function

--------------------

global function PrintCharStartPos(integer printpos, sequence rawtext)
-- get the start pos of the char whose space includes pos

    integer len, junk, old

    len = 0

    for i = 1 to length(rawtext) do
        old = len
        junk = rawtext[i]
        if junk = TAB then
            len += tabwidth - remainder(len, tabwidth)
        else
            len += 1
        end if
        if printpos <= len then
            return old + 1
        end if
    end for

    return len + 1

end function

--------------------

global function grabWord(integer col, sequence rawtext)
-- returns the print start & finish coords of the word intersected by col
    integer start, finish, len

    -- convert to byte position
    col = PrintToBytePos(col, rawtext)

    len = length(rawtext)

    -- exit if not 'on' the text at all
    if col > len or col < 1 then
        return {0, 0}
    end if

    -- returns true if char is label character
    if isLabelCharacter(rawtext[col]) then -- good, real word

        start = 1
        for i = col to 1 by -1 do
            if not isLabelCharacter(rawtext[i]) then
                start = i + 1
                exit
            end if
        end for

        finish = len
        for i = col to len do
            if not isLabelCharacter(rawtext[i]) then
                finish = i - 1
                exit
            end if
        end for

        start = ByteToPrintPos(start, rawtext)
        finish = ByteToPrintPos(finish + 1, rawtext) - 1
        return {start, finish}
    end if

    -- failure
    return {0, 0}

end function