without trace

procedure drawIndicator(integer id, integer X, integer Y, integer isleft)
    setPenColor(id, BrightRed)
    Y -= 1
    if isleft then
        drawRectangle(id, w32True, X, Y, X+10, Y+3)
    else
        drawRectangle(id, w32True, X, Y, X-10, Y+3)
    end if
end procedure

--------------------------------------------------------------------------
---------------- code to manage the tab bar to enable switching to/from opened files
sequence namelist, -- list of all opened files
         extentlist -- contains the boundaries of the accumulative widths for each tab

integer tabbar, totalwidth, tabheight, totalheight, namemargin,
    leftfile, rightfile
global integer selectedfile

    tabbar = create(Pixmap, "", 0, 0,0, 10, 10, 0)
    setFont(tabbar, "Arial", 10, 0)

    tabheight = 20
    totalheight = 30
    namemargin = 6

constant DARKSHADOW = getSysColor(COLOR_3DDKSHADOW),
         LIGHTSHADOW = getSysColor(COLOR_BTNSHADOW)
--------------------------------------------------------------------------

global procedure transferTabbarToWindow(integer id)
    if length(namelist) < 2 then return end if
    bitBlt( id, -- destination
            0, -- x coord of destination
            0, -- y coord of destination
            tabbar, -- source
            1, -- x coord of source
            1, -- y coord of source
            totalwidth, -- width
            totalheight, -- height
            SrcCopy ) -- raster operation
end procedure

--------------------------------------------------------------------------
global function isFileTabVisible(integer filepos)
    if length(namelist) = 0 then -- this is the minimum to display
        return w32False
    elsif length(namelist) = 1 then
        return w32True
    end if
    if leftfile <= filepos and filepos <= rightfile then
        return w32True
    end if
    return w32False
end function


function LowerToUpperBound(integer lefttab, integer righttab)
-- calculate inclusive distance between 2 tabs
    return extentlist[righttab + 1] - extentlist[lefttab] + 2
end function

procedure rebuildTabbar(integer ensurethistabseen)
    sequence filename
    integer lowerlimit, upperlimit, initialdisp, high

    -- repaint tabbar to background colour
    setPenColor(tabbar, mainwindowbackcolour)
    drawRectangle(tabbar, 1, 0, 0, totalwidth+1, totalheight+1)

    -- exit if less than 2 are open
    if length(namelist) < 2 then
        return
    end if

    -- determine correct left & right files tabs to show on screen
    -- (i)
    lowerlimit = ensurethistabseen
    upperlimit = lowerlimit
    for i = ensurethistabseen to leftfile by -1 do
        if LowerToUpperBound(i, upperlimit) > totalwidth then
            exit
        end if
        lowerlimit = i
    end for
    -- (ii)
    for i = ensurethistabseen+1 to length(namelist) do
        if LowerToUpperBound(lowerlimit, i) > totalwidth then
            exit
        end if
        upperlimit = i
    end for
    -- (iii)
    for i = lowerlimit-1 to 1 by -1 do
        if LowerToUpperBound(i, upperlimit) > totalwidth then
            exit
        end if
        lowerlimit = i
    end for
    -- (iv)	
    leftfile = lowerlimit
    rightfile = upperlimit

    initialdisp = extentlist[leftfile] - 2

    for i=leftfile to rightfile do

        lowerlimit = extentlist[i] - initialdisp
        upperlimit = extentlist[i+1] - 1 - initialdisp

        filename = namelist[i]
        if getStatusValue(i, SAVEFLAG) = w32True then
            filename&='*'
        end if
        if i != selectedfile then -- draw standard tab

            -- DRAW STANDARD TAB BORDER
            -- left border
            setPenColor(tabbar, BrightWhite)
            drawLine(tabbar, lowerlimit, 5, lowerlimit, tabheight+2)
            -- top border
            drawLine(tabbar, lowerlimit + 2, 3, upperlimit - 1, 3)
            setPixel(tabbar, lowerlimit + 1, 4, BrightWhite)
            setPixel(tabbar, upperlimit - 1, 4, BrightWhite)
            -- right border
            setPenColor(tabbar, Gray)
            drawLine(tabbar, upperlimit, 5, upperlimit, tabheight)
            -- set file name in the tab
            setPenColor(tabbar, Black)
            setPenPos (tabbar, lowerlimit + namemargin, 5)
            wPuts(tabbar, filename)
        end if

    end for

    -- DRAW DARK LINE OF RIGHTMOST SIDE
    if selectedfile != rightfile then
        upperlimit = extentlist[rightfile+1] - 1 - initialdisp
        setPixel(tabbar, upperlimit - 1, 4, LIGHTSHADOW)
        setPenColor(tabbar, LIGHTSHADOW)
        drawLine(tabbar, upperlimit, 5, upperlimit, tabheight+1)
    end if

    -- show selected file name if that is uppermost
    if leftfile <= selectedfile and selectedfile <= rightfile then
        lowerlimit = extentlist[selectedfile] - 1 - initialdisp
        upperlimit = extentlist[selectedfile + 1] - initialdisp
        setPenColor(tabbar, BrightWhite)
        drawLine(tabbar, lowerlimit, 3, lowerlimit, tabheight+3)
        drawLine(tabbar, lowerlimit + 1, 2, lowerlimit + 1, tabheight+2)
        drawLine(tabbar, lowerlimit + 2, 1, upperlimit - 2, 1)
        drawLine(tabbar, lowerlimit + 2, 2, upperlimit - 1, 2)

        -- draw dark line on right side
        setPenColor(tabbar, DARKSHADOW)
        drawLine(tabbar, upperlimit, 3, upperlimit, tabheight + 2)

        -- draw lighter line on right side
        setPenColor(tabbar, LIGHTSHADOW)
        drawLine(tabbar, upperlimit - 1, 2, upperlimit - 1, tabheight+1)

        -- print name on tab of selected file slightly higher than others
        setPenPos (tabbar, lowerlimit + namemargin, 3)
        setPenColor(tabbar, Black)
        filename = namelist[selectedfile]
        if getStatusValue(selectedfile, SAVEFLAG) = w32True then
            filename&='*'
        end if
        wPuts(tabbar, filename)
    end if

    -- prepare useful value
    high = extentlist[rightfile+1]- initialdisp

    -- DRAW CENTRE LINE
    setPenColor(tabbar, BrightWhite)
    if isFileTabVisible(selectedfile) then -- selection is visible
        if leftfile < selectedfile then -- draw left line
            upperlimit = extentlist[selectedfile] - initialdisp
            drawLine(tabbar, 2, tabheight+1, upperlimit, tabheight+1)
            drawLine(tabbar, 1, tabheight+2, upperlimit, tabheight+2)
        end if
        if selectedfile < rightfile then -- draw right line
            lowerlimit = extentlist[selectedfile+1] + 1 - initialdisp
            drawLine(tabbar, lowerlimit, tabheight+1, high - 1, tabheight+1)
            drawLine(tabbar, lowerlimit, tabheight+2, high - 2, tabheight+2)
            setPixel(tabbar, high - 1, tabheight+2, Gray)
        end if
    else -- centre line runs straight through
        drawLine(tabbar, 1, tabheight+1, high - 1, tabheight+1)
        drawLine(tabbar, 0, tabheight+2, high - 2, tabheight+2)
        setPixel(tabbar, high - 1, tabheight+2, Gray)
    end if

    -- DRAW ARROW BAR BORDER
    -- left line
    setPenColor(tabbar, BrightWhite)
    drawLine(tabbar, 1, tabheight+2, 1, totalheight-1)
    drawLine(tabbar, 2, tabheight+1, 2, totalheight-2)
    setPixel(tabbar, 2, totalheight-1, Gray)
    -- bottom line
    setPenColor(tabbar, DARKSHADOW)
    drawLine(tabbar, 2, totalheight, high-1, totalheight)
    setPenColor(tabbar, LIGHTSHADOW)
    drawLine(tabbar, 3, totalheight-1, high, totalheight-1)
    -- right line
    setPenColor(tabbar, DARKSHADOW)
    drawLine(tabbar, high, totalheight-1, high, tabheight+1)
    setPenColor(tabbar, LIGHTSHADOW)
    drawLine(tabbar, high-1, totalheight, high-1, tabheight)
    setPixel(tabbar, high, totalheight, LIGHTSHADOW)

    -- PRINT THE ARROW BUTTONS
    -- draw left facing arrow
    if leftfile > 1 then
        drawIndicator(tabbar, 5, tabheight+floor((totalheight-tabheight+1)/2), 1)
    end if
    -- draw right facing arrow
    if rightfile < length(namelist) then
        drawIndicator(tabbar, high-5, tabheight+floor((totalheight-tabheight+1)/2), 0)
    end if

end procedure

global function getCSMpopupplace()
    return extentlist[selectedfile+1]-(extentlist[leftfile]-2)
end function

--------------------------------------------------------------------------

procedure resizetabbarpixmap(integer width)
    setCtlSize(tabbar, width+1, totalheight+1)
    totalwidth = width
end procedure
--------------------------------------------------------------------------

global procedure resizeTabbar(integer availablewidth)
    integer minwidth, ensurethistab

    -- are tabs even visible?
    if length(namelist) < 2 then -- this is the minimum number of tabs to display
        rebuildTabbar(leftfile)
        return
    end if

    -- ensure minimum size retained
    if isFileTabVisible(selectedfile) then
        ensurethistab = selectedfile
        minwidth = LowerToUpperBound(selectedfile, selectedfile) + 1
    else
        minwidth = LowerToUpperBound(leftfile, leftfile) + 1
        ensurethistab = leftfile
    end if
    if availablewidth < minwidth then
        totalwidth = minwidth
    else
        totalwidth = availablewidth
    end if

    -- resize pixmap
    resizetabbarpixmap(totalwidth)

    -- repaint pixmap
    rebuildTabbar(ensurethistab)
end procedure
--------------------------------------------------------------------------

global procedure initializeNamesInTabbar(sequence filenames, integer barwidth)
    integer accumulatingwidth
    object width

    resizetabbarpixmap(barwidth)

    namelist = filenames
    extentlist = {}
    if not length(namelist) then -- nothing to do, really
        leftfile = 0
        selectedfile = 0
        --rebuildTabbar(1)
    end if

    -- initialize
    accumulatingwidth = 2
    extentlist = {2}

    -- loop through all names of open files
    for i=1 to length(namelist) do

        -- calculate the width occupied by each name in the tabbar
        width = getTextExtent(tabbar, namelist[i])
        width = width[5] + namemargin * 2
        accumulatingwidth += width

        -- put the accumulating value into a list
        extentlist &= accumulatingwidth
        -- nb: item x runs from extentlist[x] to extentlist[x+1]-1

    end for

    leftfile = 1
    selectedfile = 1 -- this is the only safe number to start with
    rightfile = 1

    rebuildTabbar(leftfile)
end procedure



--------------------------------------------------------------------------

global procedure selectTab(integer filepos)
    selectedfile = filepos
    rebuildTabbar(selectedfile)
end procedure

--------------------------------------------------------------------------

global procedure addTab(sequence filename)
-- add new file name to end of extentlist & make it selected
    sequence junk
    integer width

    if not length(extentlist) then
        extentlist = {2}
    end if

    namelist = append(namelist, filename)
    selectedfile = length(namelist)

    junk = getTextExtent(tabbar, filename)
    width = junk[5] + namemargin * 2
    width += extentlist[length(extentlist)]
    extentlist &= width

    -- determine the new leftfile, if any
    leftfile = 1 -- default value
    width = extentlist[length(extentlist)]
    for i=length(extentlist)-1 to 1 by -1 do
        if width - extentlist[i] > totalwidth then
            exit -- limit found
        end if
        leftfile = i
    end for

    rebuildTabbar(selectedfile)
end procedure

--------------------------------------------------------------------------

global procedure removeTab(integer filepos)
    integer lowerlimit, upperlimit, len
    len = length(namelist)

    if len = 1 then -- only one file open
        extentlist = {}
        namelist = {}
        leftfile = 0
        rightfile = 0
        selectedfile = 0

    elsif len >= 2 then
        if filepos = leftfile then
            leftfile -= 1
            if leftfile = 0 then
                leftfile = 1
            end if
        end if
        namelist = namelist[1..filepos-1] & namelist[filepos+1..len]
        lowerlimit = extentlist[filepos]
        upperlimit = extentlist[filepos+1]
        upperlimit -= lowerlimit
        extentlist[filepos..len+1] -= upperlimit
        extentlist = extentlist[1..filepos-1] & extentlist[filepos+1..len+1]

        if selectedfile = len then -- have to move back a notch
            selectedfile -= 1
        end if

    end if

    rebuildTabbar(selectedfile)
end procedure

--------------------------------------------------------------------------

global procedure renameTab(integer filepos, sequence newname)
    sequence junk
    integer disp

    -- update name string
    namelist[filepos] = newname

    -- calculate size displacement between old name & new name
    junk = getTextExtent(tabbar, newname)
    disp = junk[5] + namemargin * 2
    disp -= extentlist[filepos+1] - extentlist[filepos]

    -- update extent list
    extentlist[filepos+1..length(extentlist)] += disp

    -- repaint pixmap
    rebuildTabbar(filepos)

end procedure


global procedure assertSelectTab()
    rebuildTabbar(selectedfile)
end procedure

function isTabClickPossible(integer tabx, integer taby)
    -- any name tabs visible?
    if length(namelist) >= 2 then -- this is the minimum to display
        -- check if mouse click falls onto greater tabbar area
        if tabx > 0 and tabx < totalwidth and taby > 0 and taby < totalheight then
            return w32True
        end if
    end if
    return w32False
end function

global function isClickOnArrowBar(integer tabx, integer taby)
    integer lowerlimit, upperlimit

    if isTabClickPossible(tabx, taby) and taby > tabheight then -- arrow bar hit

        lowerlimit = extentlist[leftfile]
        upperlimit = extentlist[rightfile+1]-1
        upperlimit -= lowerlimit-1

        if tabx < floor(upperlimit/3) then -- leftside hit
            if leftfile > 1 then
                leftfile -= 1
                rebuildTabbar(leftfile)
                return w32True
            end if

        elsif tabx > floor(upperlimit*2/3) then -- rightside hit
            if leftfile < length(namelist) then
                leftfile += 1
                rebuildTabbar(leftfile)
                return w32True
            end if

        end if

    end if

    return w32False

end function
--------------------------------------------------------------------------
global function isPointOnTab(integer tabx, integer taby)
    integer lowerlimit, upperlimit

    if isTabClickPossible(tabx, taby) and taby < tabheight then

        -- convert x coord in line with extentlist
        tabx += extentlist[leftfile]

        -- check which tab it fell onto
        for i=leftfile to rightfile do
            lowerlimit = extentlist[i]
            upperlimit = extentlist[i+1]-1
            if tabx > (lowerlimit + 1) and tabx < (upperlimit - 1) then
                return i
            end if
        end for

    end if

    return 0 
end function

global function isClickOnTab(integer tabx, integer taby)
-- tabx & taby are all relative to tabbar pixmap, so calling routine will have to calculate these
    integer t
    t = isPointOnTab(tabx, taby)
    if t > 0 and t != selectedfile then -- valid file found, so switch to it
        selectTab(t)
    end if
    return t -- send back position for files.e to also switch to
end function


--------------------------------------------------------------------------

global function tabbarheightdisplacement()
    if length(namelist) < 2 then -- this is the minimum number of tabs to display
        return 0
    end if
    return totalheight --+ 7 -- this 7 is the space between the tab & edit area
end function

--------------------------------------------------------------------------

global procedure shuffleTab(integer direction)
    object temp
    if direction<0 and selectedfile=1 then return end if
    if direction>0 and selectedfile=length(namelist) then return end if
    temp=namelist[selectedfile]
    renameTab(selectedfile,namelist[selectedfile+direction])
    renameTab(selectedfile+direction,temp)
    shuffleActs(direction)
    shuffleFile(direction)
    selectedfile+=direction
end procedure
--------------------------------------------------------------------------